from typing import Any, Dict, List, Mapping, Optional, Sequence, Union

import pandas as pd
from datasets.arrow_dataset import Dataset
from datasets.iterable_dataset import IterableDataset

from hyfi.utils.iolibs import IOLIBs
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)


class DSUtils:
    def __init__(self):
        pass

    @staticmethod
    def is_dataframe(data: Any) -> bool:
        """Check if data is a pandas dataframe"""
        return isinstance(data, pd.DataFrame)

    @staticmethod
    def is_dataset(data: Any) -> bool:
        """Check if data is a HuggingFace dataset"""
        return isinstance(data, Dataset or IterableDataset)

    @staticmethod
    def get_data_files(
        data_files: Optional[
            Union[str, Sequence[str], Mapping[str, Union[str, Sequence[str]]]]
        ] = None,
        data_dir: Optional[str] = None,
        split: Optional[str] = None,
        recursive: bool = True,
        use_cached: bool = False,
        verbose: bool = False,
        **kwargs,
    ) -> Union[List[str], Dict[str, List[str]]]:
        if isinstance(data_files, dict):
            return {
                name: IOLIBs.get_filepaths(
                    files,
                    data_dir,
                    recursive=recursive,
                    use_cached=use_cached,
                    verbose=verbose,
                    **kwargs,
                )
                for name, files in data_files.items()
            }
        filepaths = IOLIBs.get_filepaths(
            data_files,
            data_dir,
            recursive=recursive,
            use_cached=use_cached,
            verbose=verbose,
            **kwargs,
        )
        return {split: filepaths} if split else filepaths

    @staticmethod
    def to_datetime(
        data,
        _format: Optional[str] = None,
        _columns: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ):
        """Convert a string, int, or datetime object to a datetime object"""
        from datetime import datetime

        import pandas as pd

        if isinstance(data, datetime):
            return data
        elif isinstance(data, str):
            if _format is None:
                _format = "%Y-%m-%d"
            return datetime.strptime(data, _format)
        elif isinstance(data, int):
            return datetime.fromtimestamp(data)
        elif isinstance(data, pd.DataFrame):
            if _columns:
                if isinstance(_columns, str):
                    _columns = [_columns]
                for _col in _columns:
                    data[_col] = pd.to_datetime(data[_col], format=_format, **kwargs)
            return data
        else:
            return data

    @staticmethod
    def to_numeric(
        data,
        _columns: Optional[Union[str, Sequence[str]]] = None,
        errors: Optional[str] = "coerce",
        downcast: Optional[str] = None,
        **kwargs,
    ):
        """Convert a string, int, or float object to a float object"""
        import pandas as pd

        if isinstance(data, str):
            return float(data)
        elif isinstance(data, (int, float)) or not isinstance(data, pd.DataFrame):
            return data
        else:
            if _columns:
                if isinstance(_columns, str):
                    _columns = [_columns]
                for _col in _columns:
                    data[_col] = pd.to_numeric(data[_col], errors=errors, downcast=downcast)  # type: ignore
            return data

    @staticmethod
    def dict_to_dataframe(
        data: Dict[Any, Any],
        orient: str = "columns",
        dtype=None,
        columns=None,
    ) -> pd.DataFrame:
        """Convert a dictionary to a pandas dataframe"""
        import pandas as pd

        return pd.DataFrame.from_dict(data, orient=orient, dtype=dtype, columns=columns)  # type: ignore

    @staticmethod
    def records_to_dataframe(
        data,
        index=None,
        exclude=None,
        columns=None,
        coerce_float=False,
        nrows=None,
    ) -> pd.DataFrame:
        """Convert a list of records to a pandas dataframe"""
        import pandas as pd

        return pd.DataFrame.from_records(
            data,
            index=index,
            exclude=exclude,
            columns=columns,
            coerce_float=coerce_float,
            nrows=nrows,
        )



import os
from pathlib import Path
from typing import Optional, Union

from hyfi.composer import BaseConfig, Composer, field_validator
from hyfi.dotenv import DotEnvConfig
from hyfi.joblib import JobLibConfig
from hyfi.path.project import ProjectPathConfig

logger = Composer.getLogger(__name__)


class ProjectConfig(BaseConfig, Composer):
    """Project Config"""

    _config_name_: str = "__init__"
    _config_group_: str = "/project"
    # Project Config
    project_name: str = "hyfi"
    project_description: Optional[str] = None
    project_root: str = "."
    project_workspace_name: str = "workspace"
    global_hyfi_root: str = "."
    global_workspace_name: str = ".hyfi"
    num_workers: int = 1
    use_huggingface_hub: bool = False
    use_wandb: bool = False
    verbose: Union[bool, int] = False
    # Config Classes
    dotenv: DotEnvConfig = None
    joblib: Optional[JobLibConfig] = None
    path: ProjectPathConfig = None

    _property_set_methods_ = {
        "project_name": "set_project_name",
        "project_root": "set_project_root",
    }

    def set_project_root(self, val: Union[str, Path]):
        if not self.project_root or self.project_root != val:
            self.path.project_root = str(val)

    def set_project_name(self, val):
        if not self.project_name or self.project_name != val:
            self.path.project_name = val

    @field_validator("project_name")
    def _validate_project_name(cls, v):
        if v is None:
            raise ValueError("Project name must be specified.")
        return v

    @field_validator("verbose")
    def _validate_verbose(cls, v):
        if isinstance(v, str):
            if v.lower() in {"true", "1"}:
                v = True
            elif v.lower() in {"false", "0"}:
                v = False
            else:
                raise ValueError("verbose must be a boolean or a string of 0 or 1")
        return v

    def __init__(self, **config_kwargs):
        super().__init__(**config_kwargs)
        self.initialize()

    def initialize(self):
        logger.debug("Initializing Project Config: %s", self.project_name)
        self.dotenv = DotEnvConfig()

        self.dotenv.HYFI_PROJECT_NAME = self.project_name
        if self.project_description:
            self.dotenv.HYFI_PROJECT_DESC = self.project_description
        if self.project_root:
            self.dotenv.HYFI_PROJECT_ROOT = self.project_root
        if self.project_workspace_name:
            self.dotenv.HYFI_PROJECT_WORKSPACE_NAME = self.project_workspace_name
        if self.global_hyfi_root:
            self.dotenv.HYFI_GLOBAL_ROOT = self.global_hyfi_root
        if self.global_workspace_name:
            self.dotenv.HYFI_GLOBAL_WORKSPACE_NAME = self.global_workspace_name
        if self.num_workers:
            self.dotenv.HYFI_NUM_WORKERS = self.num_workers
        self.dotenv.HYFI_VERBOSE = self.verbose
        self.dotenv.CACHED_PATH_CACHE_ROOT = str(self.path.cache_dir / "cached_path")

        if self.joblib:
            self.joblib.init_backend()
        else:
            logger.warning("JoblibConfig not initialized")

        self.init_wandb()
        if self.use_huggingface_hub:
            self.init_huggingface_hub()

    def init_wandb(self):
        if self.path is None:
            raise ValueError("Path object not initialized")
        if self.dotenv is None:
            raise ValueError("DotEnv object not initialized")

        if not self.use_wandb:
            return
        try:
            self._init_wandb()
        except ImportError:
            logger.warning("wandb is not installed, please install it to use wandb.")

    def _init_wandb(self):
        import wandb  # type: ignore

        self.dotenv.WANDB_DIR = str(self.path.log_dir)
        project_name = self.project_name.replace("/", "-").replace("\\", "-")
        self.dotenv.WANDB_PROJECT = project_name
        notebook_name = self.path.log_dir / f"{project_name}-nb"
        notebook_name.mkdir(parents=True, exist_ok=True)
        self.dotenv.WANDB_NOTEBOOK_NAME = str(notebook_name)
        self.dotenv.WANDB_SILENT = str(not self.verbose)

        wandb.init(project=project_name)

    def init_huggingface_hub(self):
        """Initialize huggingface_hub"""
        try:
            from huggingface_hub import notebook_login  # type: ignore
            from huggingface_hub.hf_api import HfFolder  # type: ignore
        except ImportError:
            logger.warning(
                "huggingface_hub is not installed, please install it to use huggingface_hub."
            )
            return

        self.dotenv = DotEnvConfig()
        if (
            self.dotenv.HUGGING_FACE_HUB_TOKEN is None
            and self.dotenv.HF_USER_ACCESS_TOKEN is not None
        ):
            self.dotenv.HUGGING_FACE_HUB_TOKEN = self.dotenv.HF_USER_ACCESS_TOKEN

        local_token = HfFolder.get_token()
        if local_token is None:
            if ProjectConfig.is_notebook():
                notebook_login()
            else:
                logger.info(
                    "huggingface_hub.notebook_login() is only available in notebook,"
                    "set HUGGING_FACE_HUB_TOKEN manually"
                )

    @property
    def osenv(self):
        return os.environ

    @property
    def root_dir(self) -> Path:
        self.path.project_root = self.project_root
        return self.path.root_dir

    @property
    def workspace_dir(self) -> Path:
        self.path.project_workspace_name = self.project_workspace_name
        return self.path.workspace_dir

    @property
    def global_root_dir(self) -> Path:
        self.path.global_hyfi_root = self.global_hyfi_root
        return self.path.global_root_dir

    @property
    def global_workspace_dir(self) -> Path:
        self.path.global_workspace_name = self.global_workspace_name
        return self.path.global_workspace_dir

    def get_path(
        self,
        path_name: str,
        base_dir: Optional[Union[Path, str]] = None,
    ) -> Optional[Path]:
        """
        Get the path to a directory or file.
        """
        return self.path.get_path(path_name, base_dir=base_dir) if self.path else None



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import argparse
import logging.config
import os
import sys
from typing import List, Optional

from hydra._internal.utils import (
    _run_app,
    detect_calling_file_or_module_from_stack_frame,
    detect_calling_file_or_module_from_task_function,
    detect_task_name,
    run_and_report,
)
from hydra.types import TaskFunction

from .config import create_config_search_path

log = logging.getLogger(__name__)


def _run_hydra(
    args: argparse.Namespace,
    args_parser: argparse.ArgumentParser,
    task_function: TaskFunction,
    config_path: Optional[str],
    config_name: Optional[str],
    plugins: Optional[List[str]],
    caller_stack_depth: int = 2,
) -> None:
    from hydra._internal.hydra import Hydra
    from hydra.core.global_hydra import GlobalHydra

    if args.config_name is not None:
        config_name = args.config_name

    if args.config_path is not None:
        config_path = args.config_path

    (
        calling_file,
        calling_module,
    ) = detect_calling_file_or_module_from_task_function(task_function)
    if calling_file is None and calling_module is None:
        (
            calling_file,
            calling_module,
        ) = detect_calling_file_or_module_from_stack_frame(caller_stack_depth + 1)
    task_name = detect_task_name(calling_file, calling_module)

    abs_config_dir = os.path.abspath(args.config_dir) if args.config_dir else None
    search_path = create_config_search_path(config_path, abs_config_dir, plugins)
    # validate_config_path(config_path)
    # search_path = create_automatic_config_search_path(
    #     calling_file, calling_module, config_path
    # )

    # def add_conf_dir() -> None:
    #     if args.config_dir is not None:
    #         abs_config_dir = os.path.abspath(args.config_dir)
    #         if not os.path.isdir(abs_config_dir):
    #             raise SearchPathException(
    #                 f"Additional config directory '{abs_config_dir}' not found"
    #             )
    #         search_path.prepend(
    #             provider="command-line",
    #             path=f"file://{abs_config_dir}",
    #             anchor=SearchPathQuery(provider="schema"),
    #         )

    # run_and_report(add_conf_dir)

    # def add_hyfi_conf() -> None:
    #     path = f"pkg://{__config_module_path__}"
    #     for sp_item in search_path.get_path():
    #         if sp_item.path == path:
    #             log.debug("HyFI config path already in search path")
    #             return
    #     log.debug("Adding hyfi to the config search path")
    #     search_path.prepend(
    #         provider="hyfi",
    #         path=path,
    #     )

    # run_and_report(add_hyfi_conf)

    # def add_caller_conf() -> None:
    #     caller_config_path = get_caller_config_module_path()
    #     path = f"pkg://{caller_config_path}"
    #     for sp_item in search_path.get_path():
    #         if sp_item.path == path:
    #             log.debug("Caller config path already in search path")
    #             return
    #     log.debug("Adding %s to the config search path", caller_config_path)
    #     search_path.prepend(
    #         provider="caller",
    #         path=path,
    #     )

    # run_and_report(add_caller_conf)

    hydra = run_and_report(
        lambda: Hydra.create_main_hydra2(
            task_name=task_name, config_search_path=search_path
        )
    )

    try:
        if args.help:
            hydra.app_help(config_name=config_name, args_parser=args_parser, args=args)
            sys.exit(0)
        has_show_cfg = args.cfg is not None
        if args.resolve and (not has_show_cfg and not args.help):
            raise ValueError(
                "The --resolve flag can only be used in conjunction with --cfg or --help"
            )
        if args.hydra_help:
            hydra.hydra_help(
                config_name=config_name, args_parser=args_parser, args=args
            )
            sys.exit(0)

        num_commands = (
            args.run
            + has_show_cfg
            + args.multirun
            + args.shell_completion
            + (args.info is not None)
        )
        if num_commands > 1:
            raise ValueError(
                "Only one of --run, --multirun, --cfg, --info and --shell_completion can be specified"
            )
        if num_commands == 0:
            args.run = True

        overrides = args.overrides

        if args.run or args.multirun:
            run_mode = hydra.get_mode(config_name=config_name, overrides=overrides)
            _run_app(
                run=args.run,
                multirun=args.multirun,
                mode=run_mode,
                hydra=hydra,
                config_name=config_name,
                task_function=task_function,
                overrides=overrides,
            )
        elif args.cfg:
            run_and_report(
                lambda: hydra.show_cfg(
                    config_name=config_name,
                    overrides=args.overrides,
                    cfg_type=args.cfg,
                    package=args.package,
                    resolve=args.resolve,
                )
            )
        elif args.shell_completion:
            run_and_report(
                lambda: hydra.shell_completion(
                    config_name=config_name, overrides=args.overrides
                )
            )
        elif args.info:
            hydra.show_info(
                args.info, config_name=config_name, overrides=args.overrides
            )
        else:
            sys.stderr.write("Command not specified\n")
            sys.exit(1)
    finally:
        GlobalHydra.instance().clear()



"""
    Configuration for HyFi Pipelines
"""
from typing import Any, Callable, Dict, List, Optional, Union

from hyfi.composer import (
    BaseModel,
    Composer,
    ConfigDict,
    field_validator,
    model_validator,
)
from hyfi.utils.envs import ENVs
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)


class BaseRunConfig(BaseModel):
    """Run Configuration"""

    run: Optional[Union[str, Dict[str, Any]]] = {}
    verbose: bool = False

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="allow",
        validate_assignment=False,
        alias_generator=Composer.generate_alias_for_special_keys,
    )  # type: ignore

    @model_validator(mode="before")
    def validate_model_config_before(cls, data):
        # logger.debug("Validating model config before validating each field.")
        return Composer.replace_special_keys(Composer.to_dict(data))

    @property
    def run_config(self) -> Dict[str, Any]:
        if self.run and isinstance(self.run, str):
            return {"_target_": self.run}
        return self.run or {}

    @property
    def run_target(self) -> str:
        return self.run_config.get("_target_") or ""

    @property
    def run_kwargs(self) -> Dict[str, Any]:
        _kwargs = self.run_config.copy()
        _kwargs.pop("_target_", None)
        _kwargs.pop("_partial_", None)
        return _kwargs


class RunningConfig(BaseRunConfig):
    """Running Configuration"""

    uses: str = ""


RunningSteps = List[RunningConfig]
RunningPipelines = List[RunningConfig]
RunningTasks = List[RunningConfig]
RunningCalls = List[RunningConfig]


class PipeConfig(BaseRunConfig):
    """Pipe Configuration"""

    pipe_target: str = ""
    name: Optional[str] = ""
    desc: Optional[str] = ""
    env: Optional[Dict[str, Any]] = {}
    use_pipe_obj: bool = True
    pipe_obj_arg_name: Optional[str] = ""
    return_pipe_obj: bool = False
    # task: Optional[TaskConfig] = None

    def set_enviroment(self):
        if self.env:
            ENVs.check_and_set_osenv_vars(self.env)

    def get_pipe_func(self) -> Optional[Callable]:
        if self.pipe_target.startswith("lambda"):
            return eval(self.pipe_target)
        elif self.pipe_target:
            return Composer.partial(self.pipe_target)
        else:
            return None

    def get_run_func(self) -> Optional[Callable]:
        run_cfg = self.run_config
        run_target = self.run_target
        if run_target and run_target.startswith("lambda"):
            logger.info("Returning lambda function: %s", run_target)
            return eval(run_target)
        elif run_cfg:
            if self.pipe_obj_arg_name:
                run_cfg.pop(self.pipe_obj_arg_name)
            logger.info(
                "Returning partial function: %s with kwargs: %s", run_target, run_cfg
            )
            return Composer.partial(run_cfg)
        else:
            logger.warning("No function found for %s", self)
            return None
        # if self.run.startswith("lambda"):
        #     logger.info("Returning lambda function: %s", self.run)
        #     return eval(self.run)
        # elif self.run:
        #     kwargs = self.run_with or {}
        #     if self.pipe_obj_arg_name:
        #         kwargs.pop(self.pipe_obj_arg_name)
        #     logger.info(
        #         "Returning partial function: %s with kwargs: %s", self.run, kwargs
        #     )
        #     return Composer.partial(self.run, **kwargs)
        # else:
        #     logger.warning("No function found for %s", self)
        #     return None


class DataframePipeConfig(PipeConfig):
    columns_to_apply: Optional[Union[str, List[str]]] = []
    use_batcher: bool = True
    num_workers: int = 1

    @model_validator(mode="before")
    def _check_and_set_values(cls, data):
        num_workers = data.get("num_workers")
        if num_workers and num_workers > 1:
            data["use_batcher"] = True
        return data

    @property
    def columns(self) -> List[str]:
        if not self.columns_to_apply:
            return []
        return (
            [self.columns_to_apply]
            if isinstance(self.columns_to_apply, str)
            else self.columns_to_apply
        )


Pipes = List[PipeConfig]


class PipelineConfig(BaseRunConfig):
    """Pipeline Configuration"""

    name: Optional[str] = ""
    steps: Optional[List[Union[str, Dict]]] = []
    initial_object: Optional[Any] = None
    use_task_as_initial_object: bool = False

    @field_validator("steps", mode="before")
    def steps_to_list(cls, v):
        """
        Convert a list of steps to a list

        Args:
            cls: class to use for conversion
            v: list of steps to convert

        Returns:
            list of steps converted to
        """
        return [v] if isinstance(v, str) else Composer.to_dict(v)

    def update_configs(
        self,
        rc: Union[Dict, RunningConfig],
    ):
        """
        Update running config with values from another config

        Args:
            rc: RunningConfig to update from
        """
        # If rc is a dict or dict it will be converted to RunningConfig.
        if isinstance(rc, dict):
            rc = RunningConfig(**rc)
        self.name = rc.name or self.name
        self.desc = rc.desc or self.desc

    def get_pipes(self) -> Pipes:
        """
        Get all pipes that this task is aware of

        Args:
            task: The task to use for the pipe

        Returns:
            A list of : class : `PipeConfig` objects
        """
        pipes: Pipes = []
        self.steps = self.steps or []
        # Add pipes to the pipeline.
        for rc in get_running_configs(self.steps):
            # Add a pipe to the pipeline.
            config = getattr(self, rc.uses, None)
            if isinstance(config, dict):
                pipe = PipeConfig(**Composer.update(config, rc.model_dump()))
                # Set the task to be used for the pipe.
                # if task is not None:
                #     pipe.task = task
                pipes.append(pipe)
        return pipes


Pipelines = List[PipelineConfig]


def get_running_configs(steps: list) -> List[RunningConfig]:
    """
    Parses and returns list of running configs

    Args:
        steps: list of config to parse

    Returns:
        list of : class : `RunningConfig` objects
    """
    RCs: List[RunningConfig] = []
    # Return the list of running RCs
    if not steps:
        logger.warning("No running configs provided")
        return RCs
    # Add running config to the list of running configs.
    for rc in steps:
        # Append a running config to the RCs list.
        if isinstance(rc, str):
            RCs.append(RunningConfig(uses=rc))
        elif isinstance(rc, dict):
            RCs.append(RunningConfig(**rc))
        else:
            raise ValueError(f"Invalid running config: {rc}")
    return RCs


def run_pipe(
    obj: Any,
    config: Union[Dict, PipeConfig],
) -> Any:
    """
    Run a pipe on an object

    Args:
        obj: The object to pipe on
        config: The configuration for the pipe

    Returns:
        The result of the pipe
    """
    # Create a PipeConfig object if not already a PipeConfig.
    if not isinstance(config, PipeConfig):
        config = PipeConfig(**Composer.to_dict(config))
    pipe_fn = config.get_pipe_func()
    # Return the object that is being used to execute the pipe function.
    if pipe_fn is None:
        logger.warning("No pipe function specified")
        return obj
    # Run a pipe with the pipe_fn
    if config.verbose:
        logger.info("Running a pipe with %s", config.pipe_target)
    # Apply pipe function to each object.
    if isinstance(obj, dict):
        objs = {}
        # Apply pipe to each object.
        for no, name in enumerate(obj):
            obj_ = obj[name]

            # Apply pipe to an object.
            if config.verbose:
                logger.info(
                    "Applying pipe to an object [%s], %d/%d",
                    name,
                    no + 1,
                    len(obj),
                )

            objs[name] = pipe_fn(obj_, config)
        return objs

    return pipe_fn(obj, config)



"""
    Hydra configuration management
"""
import collections.abc
import os
import re
from enum import Enum
from typing import Any, Callable, Dict, List, Mapping, Optional, Union

import hydra
from hydra.core.global_hydra import GlobalHydra
from omegaconf import DictConfig
from pydantic import BaseModel

from hyfi.core import global_hyfi
from hyfi.core import hydra as hyfi_hydra
from hyfi.utils import UTILs

if level := os.environ.get("HYFI_LOG_LEVEL"):
    UTILs.setLogger(level)
logger = UTILs.getLogger(__name__)


class SpecialKeys(str, Enum):
    """Special keys in configs used by HyFI."""

    CALL = "_call_"
    CONFIG = "_config_"
    CONFIG_GROUP = "_config_group_"
    CONFIG_NAME = "_config_name_"
    DESCRIPTION = "_description_"
    EXEC = "_exec_"
    FUNC = "_func_"
    KWARGS = "_kwargs_"
    METHOD = "_method_"
    PARTIAL = "_partial_"
    PIPE_TARGET = "pipe_target"
    RECURSIVE = "_recursive_"
    RUN = "run"
    RUN_WITH = "run_with"
    TARGET = "_target_"
    TYPE = "_type_"
    USES = "uses"
    WITH = "with"


class ConfigGroup(BaseModel):
    config_group: str = ""
    _group_override_: str = ""
    _group_key_: str = ""
    _group_value_: str = ""
    _global_package_: bool = False

    @property
    def group_override(self):
        if not self._group_override_:
            self._split_config_group()
        return self._group_override_

    @property
    def group_key(self):
        if not self._group_key_:
            self._split_config_group()
        return self._group_key_

    @property
    def group_value(self):
        if not self._group_value_:
            self._split_config_group()
        return self._group_value_

    @property
    def global_package(self):
        return self._global_package_

    def get_overrides(
        self,
        overrides: Optional[List[str]] = None,
        root_config_name: Optional[str] = None,
        config_module: Optional[str] = None,
    ) -> List[str]:
        if overrides is None:
            overrides = []
        if self.group_key in global_hyfi.global_package_list:
            self._global_package_ = True
        # If group_key and group_value are specified in the configuration file.
        if self.group_key and self.group_value:
            config = Composer.hydra_compose(
                root_config_name=root_config_name,
                config_module=config_module,
            )
            config = Composer.select(
                config,
                key=self.group_key,
                default=None,
                throw_on_missing=False,
                throw_on_resolution_failure=False,
            )
            override = (
                self.group_override if config is not None else f"+{self.group_override}"
            )
            # Add override to overrides list.
            if override and override not in overrides:
                overrides = [override] + overrides

        return overrides

    def _split_config_group(self):
        config_group = self.config_group
        if config_group:
            group_ = config_group.split("=")
            # group_key group_value group_key group_value group_key group_value default
            if len(group_) == 2:
                group_key, group_value = group_
            else:
                group_key = group_[0]
                group_value = global_hyfi.hydra_default_config_group_value
            # remove leading slash
            if group_key.startswith("/"):
                group_key = group_key[1:]
            config_group = f"{group_key}={group_value}"
        else:
            group_key = ""
            group_value = ""
            config_group = ""
        self._group_override_ = config_group
        self._group_key_ = group_key
        self._group_value_ = group_value


class Composer(UTILs):
    """
    Compose a configuration by applying overrides
    """

    @staticmethod
    def hydra_compose(
        root_config_name: Optional[str] = None,
        config_module: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        plugins: Optional[List[str]] = None,
    ):
        is_initialized = GlobalHydra.instance().is_initialized()  # type: ignore
        config_module = config_module or global_hyfi.config_module
        plugins = plugins or global_hyfi.plugins
        # logger.debug("config_module: %s", config_module)
        if is_initialized:
            # Hydra is already initialized.
            # logger.debug("Hydra is already initialized")
            cfg = hydra.compose(config_name=root_config_name, overrides=overrides)
        else:
            with hyfi_hydra.initialize_config(
                config_module=config_module,
                config_dir=global_hyfi.user_config_path,
                plugins=plugins,
                version_base=global_hyfi.hydra_version_base,
            ):
                cfg = hydra.compose(config_name=root_config_name, overrides=overrides)
        return cfg

    @staticmethod
    def compose_as_dict(
        config_group: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        config_data: Optional[Union[Dict[str, Any], DictConfig]] = None,
        throw_on_compose_failure: bool = True,
        throw_on_resolution_failure: bool = True,
        throw_on_missing: bool = False,
        root_config_name: Optional[str] = None,
        config_module: Optional[str] = None,
        global_package: bool = False,
        **kwargs,
    ) -> Dict:
        return Composer.to_dict(
            Composer.compose(
                config_group=config_group,
                overrides=overrides,
                config_data=config_data,
                throw_on_compose_failure=throw_on_compose_failure,
                throw_on_resolution_failure=throw_on_resolution_failure,
                throw_on_missing=throw_on_missing,
                root_config_name=root_config_name,
                config_module=config_module,
                global_package=global_package,
                **kwargs,
            )
        )

    @staticmethod
    def compose(
        config_group: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        config_data: Optional[Union[Dict[str, Any], DictConfig]] = None,
        throw_on_compose_failure: bool = True,
        throw_on_resolution_failure: bool = True,
        throw_on_missing: bool = False,
        root_config_name: Optional[str] = None,
        config_module: Optional[str] = None,
        global_package: bool = False,
        **kwargs,
    ) -> DictConfig:
        try:
            return Composer._compose_internal(
                config_group=config_group,
                overrides=overrides,
                config_data=config_data,
                throw_on_resolution_failure=throw_on_resolution_failure,
                throw_on_missing=throw_on_missing,
                root_config_name=root_config_name,
                config_module=config_module,
                global_package=global_package,
                **kwargs,
            )
        except Exception as e:
            logger.error("Error composing config: %s", e)
            if throw_on_compose_failure:
                raise e
            return DictConfig(config_data) if config_data else DictConfig({})

    @staticmethod
    def _compose_internal(
        config_group: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        config_data: Optional[Union[Dict[str, Any], DictConfig]] = None,
        throw_on_resolution_failure: bool = True,
        throw_on_missing: bool = False,
        root_config_name: Optional[str] = None,
        config_module: Optional[str] = None,
        global_package: bool = False,
        **kwargs,
    ) -> DictConfig:
        if isinstance(config_data, DictConfig):
            logger.debug("returning config_data without composing")
            return config_data
        cg = ConfigGroup(config_group=config_group)
        overrides = cg.get_overrides(
            overrides=overrides,
            root_config_name=root_config_name,
            config_module=config_module,
        )

        logger.debug(f"compose config with overrides: {overrides}")
        # Initialize hydra and return the configuration.
        cfg = Composer.hydra_compose(
            root_config_name=root_config_name,
            config_module=config_module,
            overrides=overrides,
        )
        # Add config group overrides to overrides list.
        global_package = global_package or cg.global_package
        if cg.group_key and not global_package:
            group_overrides: List[str] = []
            group_cfg = Composer.select(
                cfg,
                key=cg.group_key,
                default=None,
                throw_on_missing=False,
                throw_on_resolution_failure=False,
            )
            if config_data and group_cfg:
                group_overrides.extend(
                    f"{cg.group_key}.{k}={v}"
                    for k, v in config_data.items()
                    if isinstance(v, (str, int, float, bool)) and k in group_cfg
                )
            if group_overrides:
                overrides.extend(group_overrides)
                cfg = Composer.hydra_compose(
                    root_config_name=root_config_name,
                    config_module=config_module,
                    overrides=overrides,
                )

            # Select the group_key from the configuration.
            cfg = Composer.select(
                cfg,
                key=cg.group_key,
                default=None,
                throw_on_missing=throw_on_missing,
                throw_on_resolution_failure=throw_on_resolution_failure,
            )
        return cfg

    @staticmethod
    def is_composable(
        config_group: str,
        config_module: Optional[str] = None,
    ) -> bool:
        """
        Determines whether the input configuration object is composable.

        Args:
            config_group (str): The name of the configuration group to check.
            config_module (Optional[str], optional): The name of the configuration module to check. Defaults to None.

        Returns:
            bool: True if the configuration object is composable, False otherwise.
        """
        try:
            cfg = Composer.compose(
                config_group=config_group,
                config_module=config_module,
            )
            return cfg is not None
        except Exception as e:
            logger.error("Error composing config: %s", e)
            return False

    @staticmethod
    def is_instantiatable(cfg: Any):
        """
        Determines whether the input configuration object is instantiatable.

        Args:
            cfg (Any): The configuration object to check.

        Returns:
            bool: True if the configuration object is instantiatable, False otherwise.
        """
        return Composer.is_config(cfg) and SpecialKeys.TARGET in cfg

    @staticmethod
    def replace_special_keys(_dict: Mapping[str, Any]) -> Mapping:
        """
        Replace special keys in a dictionary.

        Args:
            _dict (Mapping[str, Any]): The dictionary to update.

        Returns:
            Mapping: The updated dictionary.
        """
        _new_dict = {}
        for k, v in _dict.items():
            key = Composer.generate_alias_for_special_keys(k)
            if isinstance(v, collections.abc.Mapping):
                _new_dict[key] = Composer.replace_special_keys(v)
            else:
                _new_dict[key] = v
        return _new_dict

    @staticmethod
    def generate_alias_for_special_keys(key: str) -> str:
        """
        Generate an alias for special keys.
        _with_ -> run
        _pipe_ -> pipe_target
        _run_ -> run

        Args:
            key (str): The special key to generate an alias for.

        Returns:
            str: The alias for the special key.
        """
        # replace the exact `with`, `pipe` with `run_with`, `run_pipe`
        key_ = re.sub(r"^with$", "run", key)
        # replace the prefix `_` with `run_`
        key_ = re.sub(r"^_with_$", "run", key_)
        key_ = re.sub(r"^_pipe_$", "pipe_target", key_)
        key_ = re.sub(r"^_run_$", "run", key_)
        return key_

    @staticmethod
    def partial(
        config: Union[str, Dict],
        *args: Any,
        **kwargs: Any,
    ) -> Callable:
        """
        Returns a callable object that is a partial version of the function or class specified in the config object.

        Args:
            config: An config object describing what to call and what params to use.
                    In addition to the parameters, the config must contain:
                    _target_ : target class or callable name (str)
                    And may contain:
                    _partial_: If True, return functools.partial wrapped method or object
                                False by default. Configure per target.
            args: Optional positional parameters pass-through
            kwargs: Optional named parameters to override
                    parameters in the config object. Parameters not present
                    in the config objects are being passed as is to the target.
                    IMPORTANT: dataclasses instances in kwargs are interpreted as config
                                and cannot be used as passthrough
        Returns:
            A callable object that is a partial version of the function or class specified in the config object.
        """
        if isinstance(config, str):
            config = {SpecialKeys.TARGET.value: config}
        else:
            config = Composer.to_dict(config)
        if not isinstance(config, dict):
            raise ValueError("config must be a dict or a str")
        config[SpecialKeys.PARTIAL.value] = True
        rc_kwargs_ = config.pop(SpecialKeys.KWARGS, {})
        if rc_kwargs_ and kwargs:
            kwargs.update(rc_kwargs_)
        return Composer.instantiate(config, *args, **kwargs)

    @staticmethod
    def instantiate(config: Any, *args: Any, **kwargs: Any) -> Any:
        """
        Instantiates an object using the provided config object.

        Args:
            config: An config object describing what to call and what params to use.
                    In addition to the parameters, the config must contain:
                    _target_ : target class or callable name (str)
                    And may contain:
                    _args_: List-like of positional arguments to pass to the target
                    _recursive_: Construct nested objects as well (bool).
                                    False by default.
                                    may be overridden via a _recursive_ key in
                                    the kwargs
                    _convert_: Conversion strategy
                            none    : Passed objects are DictConfig and ListConfig, default
                            partial : Passed objects are converted to dict and list, with
                                    the exception of Structured Configs (and their fields).
                            all     : Passed objects are dicts, lists and primitives without
                                    a trace of OmegaConf containers
                    _partial_: If True, return functools.partial wrapped method or object
                                False by default. Configure per target.
                    _args_: List-like of positional arguments
            args: Optional positional parameters pass-through
            kwargs: Optional named parameters to override
                    parameters in the config object. Parameters not present
                    in the config objects are being passed as is to the target.
                    IMPORTANT: dataclasses instances in kwargs are interpreted as config
                                and cannot be used as passthrough

        Returns:
            if _target_ is a class name: the instantiated object
            if _target_ is a callable: the return value of the call
        """
        verbose = config.get("verbose", False)
        if not Composer.is_instantiatable(config):
            if verbose:
                logger.info("Config is not instantiatable, returning config")
            return config
        _recursive_ = config.get(SpecialKeys.RECURSIVE, False)
        if SpecialKeys.RECURSIVE not in kwargs:
            kwargs[SpecialKeys.RECURSIVE.value] = _recursive_
        if verbose:
            logger.info("instantiating %s ...", config.get(SpecialKeys.TARGET))
        return hydra.utils.instantiate(config, *args, **kwargs)

    @staticmethod
    def getsource(obj: Any) -> str:
        """
        Return the source code of the object.

        Args:
            obj: The object to get the source code of.

        Returns:
            The source code of the object as a string.

        """
        try:
            target_string = ""
            if Composer.is_config(obj):
                if SpecialKeys.TARGET in obj:
                    target_string = obj[SpecialKeys.TARGET]
            elif isinstance(obj, str):
                target_string = obj
            return UTILs.getsource(target_string) if target_string else ""
        except Exception as e:
            logger.error(f"Error getting source: {e}")
            return ""

    @staticmethod
    def viewsource(obj: Any):
        """
        Print the source code of the object.

        Args:
            obj: The object to print the source code of.

        """
        print(Composer.getsource(obj))

    @staticmethod
    def instantiate_config(
        config_group: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        config_data: Optional[Union[Dict[str, Any], DictConfig]] = None,
        global_package: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """
        Instantiates an object using the provided config group and overrides.

        Args:
            config_group: Name of the config group to compose (`config_group=name`)
            overrides: List of config groups to apply overrides to (`overrides=["override_name"]`)
            config_data: Keyword arguments to override config group values (will be converted to overrides of the form `config_group_name.key=value`)
            global_package: If True, the config assumed to be a global package
            args: Optional positional parameters pass-through
            kwargs: Optional named parameters to override
                    parameters in the config object. Parameters not present
                    in the config objects are being passed as is to the target.
                    IMPORTANT: dataclasses instances in kwargs are interpreted as config
                                and cannot be used as passthrough

        Returns:
            if _target_ is a class name: the instantiated object
            if _target_ is a callable: the return value of the call
        """
        cfg = Composer.compose(
            config_group=config_group,
            overrides=overrides,
            config_data=config_data,
            global_package=global_package,
        )
        return Composer.instantiate(cfg, *args, **kwargs)

    @staticmethod
    def print_config(
        config_group: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        config_data: Optional[Union[Dict[str, Any], DictConfig]] = None,
        global_package: bool = False,
    ):
        """
        Print the configuration

        Args:
            config_group: Name of the config group to compose (`config_group=name`)
            overrides: List of config groups to apply overrides to (`overrides=["override_name"]`)
            config_data: Keyword arguments to override config group values (will be converted to overrides of the form `config_group_name.key=value`)
            global_package: If True, the config assumed to be a global package
        """
        cfg = Composer.compose(
            config_group=config_group,
            overrides=overrides,
            config_data=config_data,
            global_package=global_package,
        )
        Composer.print(cfg)



"""This module provides utilities for loading library packages and dependencies."""
import importlib
import inspect
import os
import subprocess
import sys
from pathlib import Path
from typing import Any

from hyfi.utils.iolibs import IOLIBs
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)


class PKGs:
    @staticmethod
    def gitclone(
        url: str,
        targetdir: str = "",
        verbose: bool = False,
    ) -> None:
        """
        Clone a git repository from the specified URL.

        Args:
            url (str): The URL of the git repository to clone.
            targetdir (str, optional): The directory to clone the repository into. Defaults to "".
            verbose (bool, optional): Whether to print the output of the git command. Defaults to False.
        """
        if targetdir:
            res = subprocess.run(
                ["git", "clone", url, targetdir], stdout=subprocess.PIPE
            ).stdout.decode("utf-8")
        else:
            res = subprocess.run(
                ["git", "clone", url], stdout=subprocess.PIPE
            ).stdout.decode("utf-8")
        if verbose:
            print(res)
        else:
            logger.info(res)

    @staticmethod
    def pip(
        name: str,
        upgrade: bool = False,
        prelease: bool = False,
        editable: bool = False,
        quiet: bool = True,
        find_links: str = "",
        requirement: bool = False,
        force_reinstall: bool = False,
        verbose: bool = False,
        **kwargs,
    ) -> None:
        """
        Install a package using pip.

        Args:
            name (str): The name of the package to install.
            upgrade (bool, optional): Whether to upgrade the package if it is already installed. Defaults to False.
            prelease (bool, optional): Whether to include pre-release versions. Defaults to False.
            editable (bool, optional): Whether to install the package in editable mode. Defaults to False.
            quiet (bool, optional): Whether to suppress output. Defaults to True.
            find_links (str, optional): URL to look for packages at. Defaults to "".
            requirement (bool, optional): Whether to install from the given requirements file. Defaults to False.
            force_reinstall (bool, optional): Whether to force a reinstall of the package. Defaults to False.
            verbose (bool, optional): Whether to print the output of the pip command. Defaults to False.
            **kwargs: Additional keyword arguments to pass to pip.

        Returns:
            None
        """
        _cmd = ["pip", "install"]
        if upgrade:
            _cmd.append("--upgrade")
        if prelease:
            _cmd.append("--pre")
        if editable:
            _cmd.append("--editable")
        if quiet:
            _cmd.append("--quiet")
        if find_links:
            _cmd += ["--find-links", find_links]
        if requirement:
            _cmd.append("--requirement")
        if force_reinstall:
            _cmd.append("--force-reinstall")
        for k in kwargs:
            k = k.replace("_", "-")
            _cmd.append(f"--{k}")
        _cmd.append(name)
        if verbose:
            logger.info(f"Installing: {' '.join(_cmd)}")
        res = subprocess.run(_cmd, stdout=subprocess.PIPE).stdout.decode("utf-8")
        if verbose:
            print(res)
        else:
            logger.info(res)

    @staticmethod
    def pipi(name: str, verbose: bool = False) -> None:
        """Install a package using pip."""
        res = subprocess.run(
            ["pip", "install", name], stdout=subprocess.PIPE
        ).stdout.decode("utf-8")
        if verbose:
            print(res)
        else:
            logger.info(res)

    @staticmethod
    def pipie(name: str, verbose: bool = False) -> None:
        """Install a editable package using pip."""
        res = subprocess.run(
            ["git", "install", "-e", name], stdout=subprocess.PIPE
        ).stdout.decode("utf-8")
        if verbose:
            print(res)
        else:
            logger.info(res)

    @staticmethod
    def apti(name: str, verbose: bool = False) -> None:
        """Install a package using apt."""
        res = subprocess.run(
            ["apt", "install", name], stdout=subprocess.PIPE
        ).stdout.decode("utf-8")
        if verbose:
            print(res)
        else:
            logger.info(res)

    @staticmethod
    def load_module_from_file(name: str, libpath: str, specname: str = "") -> None:
        """Load a module from a file"""
        module_path = os.path.join(libpath, name.replace(".", os.path.sep))
        if IOLIBs.is_file(f"{module_path}.py"):
            module_path = f"{module_path}.py"
        elif IOLIBs.is_dir(module_path):
            module_path = os.path.join(module_path, "__init__.py")
        else:
            module_path = str(Path(module_path).parent / "__init__.py")

        spec = importlib.util.spec_from_file_location(name, module_path)  # type: ignore
        module = importlib.util.module_from_spec(spec)  # type: ignore
        if not specname:
            specname = spec.name
        sys.modules[specname] = module
        spec.loader.exec_module(module)

    @staticmethod
    def ensure_import_module(
        name: str,
        libpath: str,
        liburi: str,
        specname: str = "",
        syspath: str = "",
    ) -> None:
        """Ensure a module is imported, if not, clone it from a git repo and load it"""
        try:
            if specname:
                importlib.import_module(specname)
            else:
                importlib.import_module(name)
            logger.info(f"{name} imported")
        except ImportError:
            if not os.path.exists(libpath):
                logger.info(f"{libpath} not found, cloning from {liburi}")
                PKGs.gitclone(liburi, libpath)
            if not syspath:
                syspath = libpath
            if syspath not in sys.path:
                sys.path.append(syspath)
            PKGs.load_module_from_file(name, syspath, specname)
            specname = specname or name
            logger.info(f"{name} not imported, loading from {syspath} as {specname}")

    @staticmethod
    def getsource(obj: str) -> str:
        """
        Return the source code of the object.

        Args:
            obj (str): The object to get the source code from.

        Returns:
            str: The source code of the object.

        """
        try:
            mod_name, object_name = obj.rsplit(".", 1)
            mod = importlib.import_module(mod_name)
            obj_ = getattr(mod, object_name)
            return inspect.getsource(obj_)
        except Exception as e:
            logger.error(f"Error getting source: {e}")
            return ""

    @staticmethod
    def viewsource(obj: str) -> None:
        """Print the source code of the object."""
        print(PKGs.getsource(obj))

    @staticmethod
    def get_caller_module_name(caller_stack_depth: int = 2) -> str:
        """Get the name of the module that called this function."""
        try:
            _stack = inspect.stack()
            if len(_stack) < caller_stack_depth + 1:
                logger.info(
                    "Returning top level module name (depth %d)", len(_stack) - 1
                )
                return inspect.getmodule(_stack[-1][0]).__name__  # type: ignore
            return inspect.getmodule(_stack[caller_stack_depth][0]).__name__  # type: ignore
        except Exception as e:
            logger.error(
                f"Error getting caller module name at depth {caller_stack_depth}: {e}"
            )
            return ""

    @staticmethod
    def is_importable(module_name: str) -> bool:
        module_spec = importlib.util.find_spec(module_name)  # type: ignore
        return module_spec is not None

    @staticmethod
    def safe_import_module(module_name: str) -> Any:
        """Safely imports a module."""
        try:
            return importlib.import_module(module_name)
        except ImportError:
            logger.debug("Failed to import module: %s", module_name)
            return None



from pathlib import Path
from typing import Optional, Union

from hyfi.composer import BaseConfig
from hyfi.path.dirnames import DirnamesConfig
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)


class BasePathConfig(BaseConfig):
    _config_name_: str = "__init__"
    _config_group_: str = "/path"

    dirnames: DirnamesConfig = DirnamesConfig()

    @property
    def name(self) -> str:
        """
        Returns the name of the path configuration.
        """
        return self._config_name_

    @property
    def root_dir(self) -> Path:
        """
        Returns the path to the root directory.
        """
        # return as an path
        path_ = Path(".")
        path_.mkdir(parents=True, exist_ok=True)
        return path_

    @property
    def workspace_dir(self) -> Path:
        """
        Returns the path to the workspace directory.
        """
        path_ = self.root_dir / "workspace"
        path_.mkdir(parents=True, exist_ok=True)
        return path_

    def get_path(
        self,
        path_name: str,
        base_dir: Optional[Union[Path, str]] = None,
    ) -> Path:
        """
        Get the path to a directory or file.
        """
        if not hasattr(self.dirnames, path_name):
            raise AttributeError(f"Path '{path_name}' does not exist.")
        base_dir = base_dir or self.workspace_dir
        path_ = Path(base_dir) / getattr(self.dirnames, path_name)
        path_.mkdir(parents=True, exist_ok=True)
        return path_

    @property
    def archive_dir(self) -> Path:
        """
        Returns the directory where the archives are stored.
        """
        return self.get_path("archives")

    @property
    def input_dir(self) -> Path:
        """
        Returns the directory where the inputs are stored.
        """
        return self.get_path("inputs")

    @property
    def output_dir(self) -> Path:
        """
        Returns the directory where the outputs are stored.
        """
        return self.get_path("outputs")

    @property
    def dataset_dir(self) -> Path:
        """
        Get the path to the dataset directory.
        """
        return self.get_path("datasets")

    @property
    def model_dir(self) -> Path:
        """
        Get the directory where models are stored.
        """
        return self.get_path("models")

    @property
    def module_dir(self) -> Path:
        """
        Create and return the path to the module directory.
        """
        return self.get_path("modules")

    @property
    def library_dir(self) -> Path:
        """
        Create and return the path to the library directory.
        """
        return self.get_path("library")

    @property
    def log_dir(self):
        """
        Create and return the path to the log directory.
        """
        return self.get_path("logs")

    @property
    def cache_dir(self):
        """
        Create and return the directory where cache files are stored.
        """
        return self.get_path("cache")

    @property
    def tmp_dir(self):
        """
        Create and return the directory where temporary files are stored.
        """
        return self.get_path("tmp")

    @property
    def config_dir(self):
        """
        Directory for the configuration files.
        """
        return self.get_path("configs")

    @property
    def config_filename(self):
        """
        Name of the YAML configuration file.
        """
        return f"{self.name}_{self.dirnames.config_yaml}"

    @property
    def config_jsonfile(self):
        """
        Name of the JSON configuration file.
        """
        return f"{self.name}_{self.dirnames.config_json}"

    @property
    def config_filepath(self):
        """
        Path to the YAML configuration file.
        """
        return self.config_dir / self.config_filename

    @property
    def config_jsonpath(self):
        """
        Path to the JSON configuration file.
        """
        return self.config_dir / self.config_jsonfile

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    def __str__(self):
        return f"{self.__class__.__name__}({self.name})"



import os
from typing import Dict, Mapping, Optional, Sequence, Union

import datasets as hfds
import pandas as pd
from datasets.arrow_dataset import Dataset
from datasets.dataset_dict import DatasetDict, IterableDatasetDict
from datasets.download.download_config import DownloadConfig
from datasets.download.download_manager import DownloadMode
from datasets.features import Features
from datasets.iterable_dataset import IterableDataset
from datasets.splits import Split
from datasets.utils.info_utils import VerificationMode

from hyfi.utils.contexts import elapsed_timer
from hyfi.utils.logging import LOGGING

from .transform import DSTransform
from .types import DatasetType
from .utils import DSUtils

logger = LOGGING.getLogger(__name__)


class DSLoad:
    def __init__(self):
        pass

    @staticmethod
    def load_data(
        path: Optional[str] = "pandas",
        name: Optional[str] = None,
        data_dir: Optional[str] = None,
        data_files: Optional[Union[str, Sequence[str]]] = None,
        split: Optional[str] = "train",
        filetype: Optional[str] = None,
        concatenate: Optional[bool] = False,
        use_cached: bool = False,
        verbose: Optional[bool] = False,
        **kwargs,
    ) -> Union[Dict[str, pd.DataFrame], Dict[str, DatasetType]]:
        """Load data from a file or a list of files"""
        if path in ["dataframe", "df", "pandas"]:
            data = DSLoad.load_dataframes(
                data_files,
                data_dir=data_dir,
                filetype=filetype,
                split=split,
                concatenate=concatenate,
                use_cached=use_cached,
                verbose=verbose,
                **kwargs,
            )
            if data is not None:
                return data if isinstance(data, dict) else {split: data}
            else:
                return {}
        else:
            dset = DSLoad.load_dataset(
                path,
                name=name,
                data_dir=data_dir,
                data_files=data_files,
                split=split,
                **kwargs,
            )
            split = split or "train"
            if isinstance(dset, (Dataset, IterableDataset)):
                return {split: dset}
            if concatenate:
                return {split: DSTransform.concatenate_datasets(dset.values())}
            else:
                return {k: v for k, v in dset.items() if v is not None}

    @staticmethod
    def load_dataframes(
        data_files: Union[str, Sequence[str]],
        data_dir: Optional[str] = None,
        filetype: Optional[str] = None,
        split: Optional[str] = None,
        concatenate: bool = False,
        ignore_index: bool = False,
        use_cached: bool = False,
        verbose: bool = False,
        **kwargs,
    ) -> Optional[Union[Dict[str, pd.DataFrame], pd.DataFrame]]:
        """Load data from a file or a list of files"""
        if not data_files:
            logger.warning("No data_files provided")
            return {}

        filepaths = DSUtils.get_data_files(
            data_files,
            data_dir,
            split=split,
            use_cached=use_cached,
            verbose=verbose,
            **kwargs,
        )

        if isinstance(filepaths, dict):
            data = {
                name: pd.concat(
                    [
                        DSLoad.load_dataframe(
                            f, verbose=verbose, filetype=filetype, **kwargs
                        )
                        for f in files
                    ],
                    ignore_index=ignore_index,
                )
                for name, files in filepaths.items()
            }
            data = {k: v for k, v in data.items() if v is not None}
            return data
        else:
            data = {
                os.path.basename(f): DSLoad.load_dataframe(
                    f, verbose=verbose, filetype=filetype, **kwargs
                )
                for f in filepaths
            }
            data = {k: v for k, v in data.items() if v is not None}
            if len(data) == 1:
                data = list(data.values())[0]
            elif len(data) > 1:
                if concatenate or split:
                    data = pd.concat(data.values(), ignore_index=ignore_index)
            else:
                logger.warning(f"No files found for {data_files}")
                return None
            return {split: data} if split else data

    @staticmethod
    def load_dataframe(
        data_file: str,
        data_dir: Optional[str] = None,
        filetype: Optional[str] = None,
        columns: Optional[Sequence[str]] = None,
        index_col: Union[str, int, Sequence[str], Sequence[int], None] = None,
        verbose: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        """Load a dataframe from a file"""
        dtype = kwargs.pop("dtype", None)
        if isinstance(dtype, list):
            dtype = {k: "str" for k in dtype}
        parse_dates = kwargs.pop("parse_dates", False)

        if data_file.startswith("http"):
            filepath = data_file
        else:
            filepath = os.path.join(data_dir, data_file) if data_dir else data_file
            if not os.path.exists(filepath):
                logger.warning(f"File {filepath} does not exist")
                raise FileNotFoundError(f"File {filepath} does not exist")
        filetype = (
            data_file.split(".")[-1]
            if data_file.split(".")[-1] in ["csv", "tsv", "parquet"]
            else filetype
        )
        filetype = filetype or "csv"
        filetype = filetype.replace(".", "")
        if filetype not in ["csv", "tsv", "parquet"]:
            raise ValueError("`file` should be a csv or a parquet file.")
        if verbose:
            logger.info(f"Loading data from {filepath}")
        with elapsed_timer(format_time=True) as elapsed:
            if "csv" in filetype or "tsv" in filetype:
                delimiter = kwargs.pop("delimiter", "\t") if "tsv" in filetype else None
                data = pd.read_csv(
                    filepath,
                    index_col=index_col,
                    dtype=dtype,
                    parse_dates=parse_dates,
                    delimiter=delimiter,
                )
            elif "parquet" in filetype:
                engine = kwargs.pop("engine", "pyarrow")
                if engine not in ["pyarrow", "fastparquet"]:
                    engine = "auto"
                data = pd.read_parquet(filepath, engine=engine)  # type: ignore
            else:
                raise ValueError("filetype must be .csv or .parquet")
            if isinstance(columns, list):
                columns = [c for c in columns if c in data.columns]
                data = data[columns]
            if verbose:
                logger.info(" >> elapsed time to load data: %s", elapsed())
        return data

    @staticmethod
    def load_dataset(
        path: str,
        name: Optional[str] = None,
        data_dir: Optional[str] = None,
        data_files: Optional[
            Union[str, Sequence[str], Mapping[str, Union[str, Sequence[str]]]]
        ] = None,
        split: Optional[Union[str, Split]] = None,
        cache_dir: Optional[str] = None,
        features: Optional[Features] = None,
        download_config: Optional[DownloadConfig] = None,
        download_mode: Optional[Union[DownloadMode, str]] = None,
        verification_mode: Optional[Union[VerificationMode, str]] = None,
        num_proc: Optional[int] = None,
        **config_kwargs,
    ) -> Union[DatasetDict, Dataset, IterableDatasetDict, IterableDataset]:
        """Load a dataset from the Hugging Face Hub, or a local dataset.

        You can find the list of datasets on the [Hub](https://huggingface.co/datasets) or with [`datasets.list_datasets`].

        A dataset is a directory that contains:

        - some data files in generic formats (JSON, CSV, Parquet, text, etc.).
        - and optionally a dataset script, if it requires some code to read the data files. This is used to load any kind of formats or structures.

         Returns:
            [`Dataset`] or [`DatasetDict`]:
            - if `split` is not `None`: the dataset requested,
            - if `split` is `None`, a [`~datasets.DatasetDict`] with each split.

            or [`IterableDataset`] or [`IterableDatasetDict`]: if `streaming=True`

            - if `split` is not `None`, the dataset is requested
            - if `split` is `None`, a [`~datasets.streaming.IterableDatasetDict`] with each split.

        """
        return hfds.load_dataset(
            path=path,
            name=name,
            data_dir=data_dir,
            data_files=data_files,
            split=split,
            cache_dir=cache_dir,
            features=features,
            download_config=download_config,
            download_mode=download_mode,
            verification_mode=verification_mode,
            num_proc=num_proc,
            **config_kwargs,
        )

    @staticmethod
    def load_dataset_from_disk(
        dataset_path: str,
        keep_in_memory: Optional[bool] = None,
        storage_options: Optional[dict] = None,
        num_heads: Optional[int] = 1,
        num_tails: Optional[int] = 1,
        verbose: bool = False,
    ) -> Union[Dataset, DatasetDict]:
        """Load a dataset from the filesystem."""
        data = hfds.load_from_disk(
            dataset_path=dataset_path,
            keep_in_memory=keep_in_memory,
            storage_options=storage_options,
        )
        logger.info("Dataset loaded from %s.", dataset_path)
        if verbose:
            if isinstance(data, DatasetDict or IterableDatasetDict):
                for split in data:
                    logger.info("Split: %s", split)
                    logger.info("Dataset features: %s", data[split].features)
                    logger.info("Number of records: %s", len(data[split]))
            else:
                if num_heads:
                    num_heads = min(num_heads, len(data))
                    print(data[:num_heads])
                if num_tails:
                    num_tails = min(num_tails, len(data))
                    print(data[-num_tails:])
                logger.info("Dataset features: %s", data.features)
                logger.info("Number of records: %s", len(data))

        return data



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
import functools
import inspect
from enum import Enum
from pathlib import Path, PosixPath, WindowsPath
from typing import Any, Callable, Dict, FrozenSet, Optional, TypeVar

from typing_extensions import Final, ParamSpec

from hyfi.core import global_hyfi
from hyfi.utils.logging import LOGGING

from .composer import Composer

logger = LOGGING.getLogger(__name__)

NoneType = type(None)

# Hydra-specific fields
TARGET_FIELD_NAME: Final[str] = "_target_"
PARTIAL_FIELD_NAME: Final[str] = "_partial_"
RECURSIVE_FIELD_NAME: Final[str] = "_recursive_"
CONVERT_FIELD_NAME: Final[str] = "_convert_"
POS_ARG_FIELD_NAME: Final[str] = "_args_"
DEFAULTS_LIST_FIELD_NAME: Final[str] = "defaults"

_names = [
    TARGET_FIELD_NAME,
    RECURSIVE_FIELD_NAME,
    CONVERT_FIELD_NAME,
    POS_ARG_FIELD_NAME,
    PARTIAL_FIELD_NAME,
]


HYDRA_FIELD_NAMES: FrozenSet[str] = frozenset(_names)
# Indicates types of primitive values permitted in configs
HYDRA_SUPPORTED_PRIMITIVES = frozenset(
    {
        int,
        float,
        bool,
        str,
        list,
        tuple,
        dict,
        NoneType,
        bytes,
        Path,
        PosixPath,
        WindowsPath,
    }
)

_T = TypeVar("_T")
Importable = TypeVar("Importable", bound=Callable[..., Any])
P = ParamSpec("P")
R = TypeVar("R")


_builtin_function_or_method_type = type(len)
# fmt: off
_lru_cache_type = type(functools.lru_cache(maxsize=128)(lambda: None))  # pragma: no branch
# fmt: on

_BUILTIN_TYPES: Final = (_builtin_function_or_method_type, _lru_cache_type)

del _lru_cache_type
del _builtin_function_or_method_type


class PipeTargetTypes(str, Enum):
    GENERAL_EXTERNAL_FUNCS = "__general_external_funcs__"
    GENERAL_INSTANCE_METHODS = "__general_instance_methods__"
    DATAFRAME_EXTERNAL_FUNCS = "__dataframe_external_funcs__"
    DATAFRAME_INSTANCE_METHODS = "__dataframe_instance_methods__"


class GENERATOR:
    """
    Generates Hydra configs for functions and classes.
    """

    @staticmethod
    def generate_pipe_config(
        target: Callable,
        pipe_target_type: PipeTargetTypes = PipeTargetTypes.GENERAL_EXTERNAL_FUNCS,
        use_pipe_obj: bool = True,
        pipe_obj_arg_name: Optional[str] = None,
        return_pipe_obj: bool = False,
        pipe_prefix: Optional[str] = None,
        config_name: Optional[str] = None,
        config_root: Optional[str] = None,
        **kwargs_for_target,
    ) -> str:
        """
        Generates HyFI pipe config for a given target.

        Args:
            target: Target function or class.
            pipe_target_type: Type of target function or class.
            use_pipe_obj: Whether to use pipe object as the first argument.
            pipe_obj_arg_name: Name of the pipe object argument.
            return_pipe_obj: Whether to return pipe object.
            pipe_prefix: Prefix for pipe object argument.
            config_name: Name of the config.
            config_root: Root of the config.
            **kwargs_for_target: Keyword arguments for the target.
        """
        use_first_arg_as_pipe_obj = not pipe_obj_arg_name and use_pipe_obj
        config_name = config_name or target.__name__
        config_name = f"{pipe_prefix}_{config_name}" if pipe_prefix else config_name
        config_root = config_root or global_hyfi.config_root

        run_config_name = GENERATOR.generate_config(
            target,
            use_first_arg_as_pipe_obj=use_first_arg_as_pipe_obj,
            config_name=config_name,
            config_path="run",
            config_root=config_root,
            **kwargs_for_target,
        )

        cfg = {
            "defaults": [
                pipe_target_type.value,
                {"/run": run_config_name},
            ],
            "use_pipe_obj": use_pipe_obj,
            "pipe_obj_arg_name": pipe_obj_arg_name,
            "return_pipe_obj": return_pipe_obj,
        }

        filename = f"{config_name}.yaml"
        config_path = Path(config_root) / "pipe"
        config_path.mkdir(parents=True, exist_ok=True)
        config_path /= filename
        Composer.save(cfg, config_path)
        logger.info(f"Saved HyFI pipe config for {target.__name__} to {config_path}")

        return config_name

    @staticmethod
    def generate_config(
        target: Callable,
        use_first_arg_as_pipe_obj: bool = False,
        config_name: Optional[str] = None,
        config_path: str = "run",
        config_root: Optional[str] = None,
        **kwargs_for_target,
    ) -> str:
        """
        Saves a HyFI config to a file.

        Args:
            target (Callable): The function or class to generate a config for.
            use_first_arg_as_pipe_obj (bool): Whether to use the first argument as the pipe object.
            config_name (Optional[str]): The name of the config. If not provided, the name of the target will be used.
            config_path (Optional[str]): The path to save the config to (relative to the config root). Defaults to "run".
            config_root (Optional[str]): The root of the config path. If not provided, the global hyfi config directory will be used.
            **kwargs_for_target: Keyword arguments to pass to the target.
        """
        cfg = GENERATOR.generate_hyfi_config(
            target,
            remove_first_arg=use_first_arg_as_pipe_obj,
            **kwargs_for_target,
        )
        config_name = config_name or target.__name__
        filename = f"{config_name}.yaml"
        config_root = config_root or global_hyfi.config_root
        config_path = Path(config_root) / config_path
        config_path.mkdir(parents=True, exist_ok=True)
        config_path /= filename
        Composer.save(cfg, config_path)
        logger.info(f"Saved HyFI config for {target.__name__} to {config_path}")
        return config_name

    @staticmethod
    def generate_hyfi_config(
        target: Callable,
        remove_first_arg: bool = False,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Generates a HyFI config for a given target.

        Args:
            target (Callable): The function or class to generate a config for.
            remove_first_arg (bool): Whether to remove the first argument from the config.
            **kwargs: Keyword arguments to pass to the target.
        """
        params = inspect.signature(target).parameters

        config_dict = {
            "_target_": f"{target.__module__}.{target.__qualname__}",
        }

        for i, (key, param) in enumerate(params.items()):
            if remove_first_arg and i == 0:
                continue
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                continue
            if key in kwargs:
                value = kwargs[key]
            else:
                value = (
                    None if param.default == inspect.Parameter.empty else param.default
                )
            config_dict[key] = sanitized_default_value(value)
        return config_dict


def sanitized_default_value(
    value: Any,
) -> Any:
    """
    Converts `value` to Hydra-supported type if necessary and possible.
    Otherwise return None.
    """
    # Common primitives supported by Hydra.
    # We check exhaustively for all Hydra-supported primitives below but seek to
    # speedup checks for common types here.
    if value is None or type(value) in {str, int, bool, float}:
        return value

    # non-str collection
    if hasattr(value, "__iter__"):
        return sanitize_collection(value)

    # basemodel with _config_name_ attribute
    # if hasattr(value, "_config_name_"):
    #     return value._config_name_

    # importable callable (function, type, or method)
    if callable(value) and (
        inspect.isfunction(value)
        or inspect.isclass(value)
        or inspect.ismethod(value)
        or isinstance(value, _BUILTIN_TYPES)
    ):
        # `value` is importable callable -- create config that will import
        # `value` upon instantiation
        return GENERATOR.generate_hyfi_config(value)

    return None


def sanitize_collection(
    x: _T,
) -> _T:
    """Pass contents of lists, tuples, or dicts through sanitized_default_values"""
    type_x = type(x)
    if type_x in {list, tuple}:
        return type_x(sanitized_default_value(_x) for _x in x)
    elif type_x is dict:
        return {
            sanitized_default_value(k): sanitized_default_value(v) for k, v in x.items()
        }
    else:
        return None



from pathlib import Path

from hyfi.utils.logging import LOGGING

from .task import TaskPathConfig

logger = LOGGING.getLogger(__name__)


class BatchPathConfig(TaskPathConfig):
    _config_name_: str = "__batch__"

    batch_name: str = "demo-batch"

    @property
    def batch_dir(self) -> Path:
        """
        Returns the path to the batch directory.
        """
        path_ = self.output_dir / self.batch_name
        path_.mkdir(parents=True, exist_ok=True)
        return path_



"""
    This module contains the primary class for the hyfi config package, HyFI,
    as well as various utility functions and imports.
"""
import os
import random
from typing import Any, Dict, List, Optional, Union

import hydra
from omegaconf import DictConfig, OmegaConf

from hyfi.cached_path import cached_path
from hyfi.composer import (
    GENERATOR,
    BaseModel,
    Composer,
    ConfigDict,
    FieldValidationInfo,
    field_validator,
)
from hyfi.copier import Copier
from hyfi.core import GlobalHyFIResolver, global_hyfi
from hyfi.dotenv import DotEnvConfig
from hyfi.graphics import GRAPHICs
from hyfi.joblib import BATCHER, JobLibConfig
from hyfi.pipeline import PIPELINEs
from hyfi.project import ProjectConfig
from hyfi.task import TaskConfig
from hyfi.workflow import WorkflowConfig

from .config import GlobalConfigResolver, global_config

logger = Composer.getLogger(__name__)

ConfigType = Union[DictConfig, Dict]

OmegaConf.register_new_resolver("__hyfi_path__", GlobalHyFIResolver.__hyfi_path__)
OmegaConf.register_new_resolver("__hyfi_version__", GlobalHyFIResolver.__hyfi_version__)
OmegaConf.register_new_resolver("__package_name__", GlobalHyFIResolver.__package_name__)
OmegaConf.register_new_resolver("__package_path__", GlobalHyFIResolver.__package_path__)
OmegaConf.register_new_resolver("__app_version__", GlobalHyFIResolver.__app_version__)
OmegaConf.register_new_resolver("__version__", GlobalHyFIResolver.__app_version__)
OmegaConf.register_new_resolver(
    "__config_module_path__", GlobalHyFIResolver.__config_module_path__
)
OmegaConf.register_new_resolver(
    "__user_config_path__", GlobalHyFIResolver.__user_config_path__
)
OmegaConf.register_new_resolver("__home_path__", GlobalHyFIResolver.__home_path__)
OmegaConf.register_new_resolver(
    "__project_root_path__", GlobalConfigResolver.__project_root_path__
)
OmegaConf.register_new_resolver(
    "__project_workspace_path__", GlobalConfigResolver.__project_workspace_path__
)
OmegaConf.register_new_resolver("__get_path__", GlobalConfigResolver.__get_path__)
OmegaConf.register_new_resolver("today", Composer.today)
OmegaConf.register_new_resolver("to_datetime", Composer.strptime)
OmegaConf.register_new_resolver("iif", lambda cond, t, f: t if cond else f)
OmegaConf.register_new_resolver("alt", lambda val, alt: val or alt)
OmegaConf.register_new_resolver("randint", random.randint, use_cache=True)
OmegaConf.register_new_resolver("get_method", hydra.utils.get_method)
OmegaConf.register_new_resolver("get_original_cwd", Composer.getcwd)
OmegaConf.register_new_resolver("exists", Composer.exists)
OmegaConf.register_new_resolver("join_path", Composer.join_path)
OmegaConf.register_new_resolver("mkdir", Composer.mkdir)
OmegaConf.register_new_resolver("dirname", os.path.dirname)
OmegaConf.register_new_resolver("basename", os.path.basename)
OmegaConf.register_new_resolver("check_path", Composer.check_path)
OmegaConf.register_new_resolver("cached_path", cached_path)
OmegaConf.register_new_resolver(
    "lower_case_with_underscores", Composer.lower_case_with_underscores
)
OmegaConf.register_new_resolver("dotenv_values", Composer.dotenv_values)


class HyFI(
    BaseModel,
    BATCHER,
    Composer,
    GENERATOR,
    GRAPHICs,
    PIPELINEs,
):
    """Primary class for the hyfi config package"""

    _config_name_: str = "config"
    _config_group_: str = "/"

    debug_mode: bool = False
    noop: bool = False
    dryrun: bool = False
    resolve: bool = False
    verbose: bool = False
    logging_level: str = "WARNING"

    hydra: Optional[ConfigType] = None

    about: Optional[ConfigType] = None
    copier: Optional[ConfigType] = None
    project: Optional[ConfigType] = None
    pipeline: Optional[ConfigType] = None
    task: Optional[ConfigType] = None
    workflow: Optional[ConfigType] = None
    tasks: Optional[List[str]] = None
    pipelines: Optional[List[str]] = None

    __version__ = GlobalHyFIResolver.__hyfi_version__()
    __hyfi_path__ = GlobalHyFIResolver.__hyfi_path__()
    __home_path__ = GlobalHyFIResolver.__home_path__()
    __package_name__ = GlobalHyFIResolver.__package_name__()
    __package_path__ = GlobalHyFIResolver.__package_path__()
    __app_version__ = GlobalHyFIResolver.__app_version__()

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
        extra="allow",
    )  # type: ignore

    @field_validator("logging_level")
    def _validate_logging_level(cls, v, info: FieldValidationInfo):
        """
        Validate and set the logging level
        """
        verbose = info.data.get("verbose", False)
        # Set verbose to INFO.
        if verbose and v == "WARNING":
            v = "INFO"
        logger.setLevel(v)
        return v

    @property
    def app_name(self):
        """
        Get the name of the application.

        Returns:
            The name of the application
        """
        return self.about.name if self.about else global_hyfi.hyfi_name

    @staticmethod
    def initialize_global_hyfi(
        package_path: str,
        version: str,
        plugins: Optional[List[str]] = None,
        user_config_path: Optional[str] = None,
        config_dirname: Optional[str] = None,
        dotenv_file: Optional[str] = None,
        secrets_dir: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Initializes the global HyFI instance.

        This function should be called before any other HyFI function.

        A plugin is a python module which contains a configuration module.

        Be careful!
        It does not check if the plugin is importable.

        Args:
            package_path: Path to the package root folder. e.g. `./src/hyfi`
            version: Version of the package. e.g. `0.1.0`
            plugins: A list of plugins to load. e.g. `["hyfi.conf"]`
            user_config_path: Path to the user configuration directory. e.g. `./config`
            config_dirname: Name of the configuration directory. e.g. `conf`
            dotenv_file: Name of the dotenv file. e.g. `.env`
            secrets_dir: Name of the secrets directory. e.g. `secrets`
            **kwargs: Additional arguments to be set as attributes.
        """
        global_hyfi.initialize(
            package_path=package_path,
            version=version,
            plugins=plugins,
            user_config_path=user_config_path,
            config_dirname=config_dirname,
            dotenv_file=dotenv_file,
            secrets_dir=secrets_dir,
            **kwargs,
        )

    @staticmethod
    def print_about(**args) -> None:
        """Print the about information"""
        global_config.print_about(**args)

    @staticmethod
    def init_project(
        project_name: Optional[str] = None,
        project_description: Optional[str] = None,
        project_root: Optional[str] = None,
        project_workspace_name: Optional[str] = None,
        global_hyfi_root: Optional[str] = None,
        global_workspace_name: Optional[str] = None,
        num_workers: Optional[int] = None,
        log_level: Optional[str] = None,
        reinit: bool = True,
        autotime: bool = True,
        retina: bool = True,
        verbose: Union[bool, int] = False,
        **project_kwargs,
    ) -> ProjectConfig:
        """
        Initialize and start hyfi.

        Args:
            project_name: Name of the project to use.
            project_description: Description of the project that will be used.
            project_root: Root directory of the project.
            project_workspace_name: Name of the project's workspace directory.
            global_hyfi_root: Root directory of the global hyfi.
            global_workspace_name: Name of the global hierachical workspace directory.
            num_workers: Number of workers to run.
            log_level: Log level for the log.
            autotime: Whether to automatically set time and / or keep track of run times.
            retina: Whether to use retina or not.
            verbose: Enables or disables logging
        """
        global_config.init_project(
            project_name=project_name,
            project_description=project_description,
            project_root=project_root,
            project_workspace_name=project_workspace_name,
            global_hyfi_root=global_hyfi_root,
            global_workspace_name=global_workspace_name,
            num_workers=num_workers,
            log_level=log_level,
            reinit=reinit,
            autotime=autotime,
            retina=retina,
            verbose=verbose,
            **HyFI.to_dict(project_kwargs),
        )
        if global_config.project:
            return global_config.project
        else:
            raise ValueError("Project not initialized.")

    @staticmethod
    def set_project(project: ProjectConfig) -> None:
        """
        Set the project.

        Args:
            project: Project to set.
        """
        logger.info(f"Setting the global project to {project.project_name}")
        global_config.project = project

    @staticmethod
    def terminate() -> None:
        """
        Terminate the global config.

        Returns:
            bool: True if termination was successful, False otherwise.
        """
        global_config.terminate()

    @staticmethod
    def JobLibConfig(**kwargs) -> JobLibConfig:
        """
        Return the joblib pipe.

        Args:
            **kwargs: Additional keyword arguments to pass to the JobLibConfig constructor.

        Returns:
            JobLibConfig: An instance of the JobLibConfig class.
        """
        return JobLibConfig(**kwargs)

    @staticmethod
    def DotEnvConfig(**kwargs) -> DotEnvConfig:
        """
        Return the DotEnvConfig.

        Args:
            **kwargs: Additional keyword arguments to pass to the DotEnvConfig constructor.

        Returns:
            DotEnvConfig: An instance of the DotEnvConfig class.
        """
        return DotEnvConfig(**kwargs)

    @staticmethod
    def TaskConfig(**kwargs) -> TaskConfig:
        """
        Return the TaskConfig.

        Args:
            **kwargs: Additional keyword arguments to pass to the TaskConfig constructor.

        Returns:
            TaskConfig: An instance of the TaskConfig class.
        """
        return TaskConfig(**kwargs)

    @staticmethod
    def WorkflowConfig(**kwargs) -> WorkflowConfig:
        """
        Return the WorkflowConfig.

        Args:
            **kwargs: Additional keyword arguments to pass to the WorkflowConfig constructor.

        Returns:
            WorkflowConfig: An instance of the WorkflowConfig class.
        """
        return WorkflowConfig(**kwargs)

    ###############################
    # Pipeline related functions
    ###############################
    @staticmethod
    def run_command(**config):
        """Run a command"""
        return HyFI.run_config(config)

    @staticmethod
    def run(
        config_group: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        config_data: Optional[Union[Dict[str, Any], DictConfig]] = None,
        global_package=False,
        dryrun=False,
        **kwargs,
    ):
        """Run the config by composing it and running it"""
        if config_group:
            logger.info("Composing the HyFI config from config group %s", config_group)
            config = HyFI.compose_as_dict(
                config_group=config_group,
                overrides=overrides,
                config_data=config_data,
                global_package=global_package,
            )
        HyFI.run_config(config, dryrun=dryrun)

    @staticmethod
    def run_intantiatable(
        config: Dict[str, Any],
        dryrun=False,
    ):
        """Run the config by composing it and running it"""
        logger.info("Instantiating the HyFI config")
        if dryrun:
            print("\nDryrun is enabled, not running the HyFI config\n")
            return
        task = HyFI.instantiate(config)
        if task and getattr(task, "__call__", None):
            logger.info("The HyFI config is callable, running it")
            task()

    @staticmethod
    def run_config(
        config: Union[Dict[str, Any], DictConfig],
        dryrun=False,
    ):
        """Run the provided config"""
        config = HyFI.to_dict(config) if config else {}
        if not isinstance(config, dict):
            raise ValueError("The config must be a dictionary")
        cmd_name = config.get("cmd_name")
        # Check if the config is instantiatable
        if HyFI.is_instantiatable(config):
            HyFI.run_intantiatable(config, dryrun=dryrun)
        else:
            logger.info(
                "The HyFI config is not instantiatable, running HyFI task with the config"
            )
            # Run the HyFI task
            config_group = config.get("_config_group_", "")
            if config_group == "/workflow" or cmd_name == "run_workflow":
                workflow = HyFI.WorkflowConfig(**config)
                HyFI.run_workflow(workflow, dryrun=dryrun)
            elif "task" in config and (cmd_name is None or cmd_name == "run_task"):
                project = (
                    HyFI.init_project(**config["project"])
                    if "project" in config
                    else None
                )
                task = HyFI.TaskConfig(**config["task"])
                HyFI.run_task(task, project=project, dryrun=dryrun)
            elif "runner" in config:
                runner = config["runner"]
                HyFI.run_intantiatable(runner, dryrun)
            elif "copier" in config and (cmd_name is None or cmd_name == "copy_conf"):
                copier_cfg = config["copier"]
                copier_cfg["dryrun"] = dryrun
                with Copier(**copier_cfg) as worker:
                    worker.run_copy()
            else:
                HyFI.print_about(**config.get("about", {}))



from .config import GlobalConfig, GlobalConfigResolver, global_config
from .main import HyFI

__all__ = [
    "HyFI",
    "GlobalConfig",
    "global_config",
    "__project_root_path__",
    "__project_workspace_path__",
]



from .pipes import (
    dataframe_external_funcs,
    dataframe_instance_methods,
    general_external_funcs,
    general_instance_methods,
)

__all__ = [
    "general_external_funcs",
    "general_instance_methods",
    "dataframe_external_funcs",
    "dataframe_instance_methods",
]



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import copy
import os
from typing import Any, List, Optional

from hydra import version
from hydra._internal.config_search_path_impl import ConfigSearchPathImpl
from hydra._internal.hydra import Hydra
from hydra.core.config_search_path import ConfigSearchPath
from hydra.core.global_hydra import GlobalHydra
from hydra.core.singleton import Singleton
from hydra.errors import HydraException

from hyfi.core import global_hyfi
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)


def get_gh_backup() -> Any:
    if GlobalHydra in Singleton._instances:
        return copy.deepcopy(Singleton._instances[GlobalHydra])
    else:
        return None


def restore_gh_from_backup(_gh_backup: Any) -> Any:
    if _gh_backup is None:
        del Singleton._instances[GlobalHydra]
    else:
        Singleton._instances[GlobalHydra] = _gh_backup


_UNSPECIFIED_: Any = object()


class initialize_config:
    """
    Initializes Hydra and add the config_module to the config search path.
    The config module must be importable (an __init__.py must exist at its top level)
    Optionally, a config_dir can be specified to add a file:// search path to it.

    Args:
        config_module: absolute module name, for example "hyfi.conf".
        config_dir: file system path to the config directory (default is None)
        job_name: the value for hydra.job.name (default is 'app')
    """

    def __init__(
        self,
        config_module: str,
        config_dir: Optional[str] = None,
        plugins: Optional[List[str]] = None,
        job_name: str = "app",
        version_base: Optional[str] = _UNSPECIFIED_,
    ) -> None:
        self._gh_backup = get_gh_backup()

        version.setbase(version_base)

        # Relative here would be interpreted as relative to cwd, which - depending on when it run
        # may have unexpected meaning. best to force an absolute path to avoid confusion.
        # Can consider using hydra.utils.to_absolute_path() to convert it at a future point if there is demand.
        if config_dir and not os.path.isabs(config_dir):
            raise HydraException(
                "initialize_config_dir() requires an absolute config_dir as input"
            )
        csp = create_config_search_path(
            config_module=config_module, search_path_dir=config_dir, plugins=plugins
        )
        Hydra.create_main_hydra2(task_name=job_name, config_search_path=csp)

    def __enter__(self, *args: Any, **kwargs: Any) -> None:
        ...

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        restore_gh_from_backup(self._gh_backup)

    def __repr__(self) -> str:
        return "hyfi.core.hydra.initialize_config()"


def append_search_path(provider: str, path: str, search_path: ConfigSearchPath) -> None:
    if not path:
        if global_hyfi.verbosity > 1:
            logger.debug(
                "Not adding empty path to Hydra's config search path for `%s`", provider
            )
        return
    for sp_item in search_path.get_path():
        if sp_item.path == path:
            if global_hyfi.verbosity > 1:
                logger.debug(
                    "Not adding `%s` to Hydra's config search path for `%s` because it is already there by `%s`",
                    path,
                    provider,
                    sp_item.provider,
                )
            return
    if global_hyfi.verbosity > 0:
        logger.debug(
            "Adding `%s` to Hydra's config search path for `%s`", path, provider
        )
    search_path.append(provider, path)


def create_config_search_path(
    config_module: Optional[str],
    search_path_dir: Optional[str],
    plugins: Optional[List[str]],
) -> ConfigSearchPath:
    from hydra.core.plugins import Plugins
    from hydra.plugins.search_path_plugin import SearchPathPlugin

    search_path = ConfigSearchPathImpl()
    search_path.append("hydra", "pkg://hydra.conf")

    # addiing hyfi's config module to the search path should come before the other modules
    append_search_path("hyfi", global_hyfi.hyfi_config_module_path, search_path)

    if plugins:
        for plugin in plugins:
            plugin_name = plugin.split(".")[0]
            plugin_name = (
                plugin_name if plugin_name.startswith("hyfi") else f"hyfi-{plugin_name}"
            )
            append_search_path(plugin_name, f"pkg://{plugin}", search_path)

    if config_module:
        path = (
            config_module
            if config_module.startswith("pkg://")
            else f"pkg://{config_module}"
            if "." in config_module
            else ""
        )
        append_search_path("main", path, search_path)

    if search_path_dir is not None and os.path.isdir(search_path_dir):
        append_search_path("user", f"file://{search_path_dir}", search_path)

    search_path_plugins = Plugins.instance().discover(SearchPathPlugin)
    for spp in search_path_plugins:
        plugin = spp()
        assert isinstance(plugin, SearchPathPlugin)
        plugin.manipulate_search_path(search_path)

    append_search_path("schema", "structured://", search_path)

    return search_path



from .core import GlobalHyFIResolver
from .core import __global_hyfi__ as global_hyfi
from .core import _batcher_instance_ as global_batcher

__all__ = ["global_hyfi", "GlobalHyFIResolver", "global_batcher"]



"""
    HyFI: Hydra Fast Interface (Hydra and Pydantic based interface framework)

    All names of configuration folders and their counterparts folders for pydantic models
    are singular, e.g. `conf` instead of `configs`.
    All matching pydantic models are named `Config`, e.g. `BatchConfig` instead of `BatchConfigs`.

    Other module folders are plural, e.g. `utils` instead of `util`.
"""
from typing import List, Optional

from hyfi.__cli__ import hydra_main, hyfi_main
from hyfi.core import global_hyfi
from hyfi.main import HyFI
from hyfi.main import HyFI as H
from hyfi.main import HyFI as HI
from hyfi.main import global_config
from hyfi.utils.logging import LOGGING

__all__ = [
    "global_config",
    "global_hyfi",
    "hydra_main",
    "hyfi_main",
    "HyFI",
    "H",
    "HI",
    "LOGGING",
]



"""
Hyfi configuration file.
"""
from pathlib import Path
from typing import Dict, List, Optional, Union

from omegaconf import DictConfig

from hyfi.about import AboutConfig
from hyfi.composer import (
    BaseModel,
    ConfigDict,
    FieldValidationInfo,
    PrivateAttr,
    field_validator,
)
from hyfi.core import global_hyfi
from hyfi.dotenv import DotEnvConfig
from hyfi.project import ProjectConfig
from hyfi.utils import UTILs

logger = UTILs.getLogger(__name__)

ConfigType = Union[DictConfig, Dict]

__default_project_root__ = "."
__default_workspace_name__ = "workspace"


class GlobalConfig(BaseModel, UTILs):
    """HyFI global config class.  This class is used to store the configuration"""

    debug_mode: bool = False
    noop: bool = False
    dryrun: bool = False
    resolve: bool = False
    verbose: bool = False
    logging_level: str = "WARNING"

    hydra: Optional[ConfigType] = None

    about: Optional[AboutConfig] = None
    copier: Optional[ConfigType] = None
    project: Optional[ProjectConfig] = None
    pipeline: Optional[ConfigType] = None
    task: Optional[ConfigType] = None
    workflow: Optional[ConfigType] = None
    tasks: Optional[List[str]] = None

    _version_: str = PrivateAttr(global_hyfi.version)
    _initilized_: bool = PrivateAttr(False)

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
        extra="allow",
    )  # type: ignore

    @field_validator("logging_level")
    def _validate_logging_level(cls, v, info: FieldValidationInfo):
        """
        Validate and set the logging level
        """
        verbose = info.data.get("verbose", False)
        # Set verbose to INFO.
        if verbose and v == "WARNING":
            v = "INFO"
        logger.setLevel(v)
        return v

    def init_project(
        self,
        project_name: Optional[str] = None,
        project_description: Optional[str] = None,
        project_root: Optional[str] = None,
        project_workspace_name: Optional[str] = None,
        global_hyfi_root: Optional[str] = None,
        global_workspace_name: Optional[str] = None,
        num_workers: Optional[int] = None,
        log_level: Optional[str] = None,
        reinit: bool = True,
        autotime: bool = True,
        retina: bool = True,
        verbose: Union[bool, int] = False,
        **project_kwargs,
    ):
        """
        Initialize and start hyfi.

        Args:
            project_name: Name of the project to use.
            project_description: Description of the project that will be used.
            project_root: Root directory of the project.
            project_workspace_name: Name of the project's workspace directory.
            global_hyfi_root: Root directory of the global hyfi.
            global_workspace_name: Name of the global hierachical workspace directory.
            num_workers: Number of workers to run.
            log_level: Log level for the log.
            autotime: Whether to automatically set time and / or keep track of run times.
            retina: Whether to use retina or not.
            verbose: Enables or disables logging
        """
        if self._initilized_ and not reinit:
            return
        if self.about is None:
            self.about = AboutConfig()
        # ENVs.load_dotenv()
        envs = DotEnvConfig(HYFI_VERBOSE=verbose)  # type: ignore
        # Set the project name environment variable HYFI_PROJECT_NAME environment variable if project_name is not set.
        if project_name:
            envs.HYFI_PROJECT_NAME = GlobalConfig.expand_posix_vars(project_name)
        # Set the project description environment variable HYFI_PROJECT_DESC environment variable.
        if project_description:
            envs.HYFI_PROJECT_DESC = GlobalConfig.expand_posix_vars(project_description)
        # Set environment variables HYFI_PROJECT_ROOT to the project root if project_root is set to true.
        if project_root:
            envs.HYFI_PROJECT_ROOT = GlobalConfig.expand_posix_vars(project_root)
        # Set the project workspace name environment variable HYFI_PROJECT_WORKSPACE_NAME environment variable if project_workspace_name is set to the project workspace name.
        if project_workspace_name:
            envs.HYFI_PROJECT_WORKSPACE_NAME = GlobalConfig.expand_posix_vars(
                project_workspace_name
            )
        # Expand the hyfi_root environment variable.
        if global_hyfi_root:
            envs.HYFI_GLOBAL_ROOT = GlobalConfig.expand_posix_vars(global_hyfi_root)
        # Set the global workspace name environment variable HYFI_GLOBAL_WORKSPACE_NAME environment variable.
        if global_workspace_name:
            envs.HYFI_GLOBAL_WORKSPACE_NAME = GlobalConfig.expand_posix_vars(
                global_workspace_name
            )
        # Set the number of workers to use.
        if num_workers:
            envs.HYFI_NUM_WORKERS = num_workers
        # Set the log level to the given log level.
        if log_level:
            envs.HYFI_LOG_LEVEL = log_level
            GlobalConfig.setLogger(log_level)
            logger.setLevel(log_level)
        # Load the extentions for the autotime extension.
        if autotime:
            GlobalConfig.load_extentions(exts=["autotime"])
        # Set the retina matplotlib formats.
        if retina:
            GlobalConfig.set_matplotlib_formats("retina")

        self.project = ProjectConfig(**project_kwargs)
        logger.info("HyFi project [%s] initialized", self.project.project_name)
        self._initilized_ = True

    def terminate(self) -> None:
        """
        Terminate hyfi config by stopping joblib

        Returns:
            True if successful False
        """
        # If the module is not initialized yet.
        if not self._initilized_:
            return
        # Stop the backend if the joblib is running.
        if self.project and self.project.joblib:
            self.project.joblib.stop_backend()
        self._initilized_ = False

    def __repr__(self):
        """
        Returns a string representation of HyFIConfig.

        Returns:
            The string representation of HyFI
        """
        return f"HyFIConfig(project={self.project})"

    def __str__(self):
        """
        Returns a string representation of the object.

        Returns:
                The string representation of the
        """
        return self.__repr__()

    @property
    def app_version(self):
        """
        Get the version of the application.


        Returns:
            The version of the application
        """
        return global_hyfi.version

    @property
    def app_name(self):
        """
        Get the name of the application.

        Returns:
            The name of the application
        """
        return self.about.name if self.about else global_hyfi.hyfi_name

    @property
    def package_name(self):
        """
        Get the name of the package.

        Returns:
            The name of the package
        """
        return global_hyfi.package_name

    def print_about(self, **kwargs):
        self.about = AboutConfig(**kwargs)
        pkg_name = self.package_name
        name = self.app_name
        print()
        for k, v in self.about.model_dump().items():
            if k.startswith("_") or k == "version":
                continue
            print(f"{k:11} : {v}")
        print(f"{'version':11} : {self.app_version}")
        if pkg_name:
            print(f"\nExecute `{pkg_name} --help` to see what you can do with {name}")

    @property
    def project_dir(self) -> Path:
        """Get the project root directory."""
        return (
            self.project.root_dir
            if self.project
            else Path(__default_project_root__).absolute()
        )

    @property
    def project_workspace_dir(self) -> Path:
        """Get the project workspace directory."""
        return (
            self.project.workspace_dir
            if self.project
            else self.project_dir / __default_workspace_name__
        )

    def get_path(
        self,
        path_name: str,
        base_dir: Optional[Union[Path, str]] = None,
    ) -> Optional[Path]:
        """
        Get the path to a directory or file.
        """
        return (
            self.project.get_path(path_name, base_dir=base_dir)
            if self.project
            else None
        )


global_config = GlobalConfig()


class GlobalConfigResolver:
    @staticmethod
    def __project_root_path__() -> str:
        """Global HyFI config path for the project root."""
        return str(global_config.project_dir or "")

    @staticmethod
    def __project_workspace_path__() -> str:
        """Global HyFI config path for the project workspace directory."""
        return str(global_config.project_workspace_dir or "")

    @staticmethod
    def __get_path__(path_name: str, base_dir: Optional[str] = None) -> str:
        """
        Get the path to a directory or file.
        """
        return str(global_config.get_path(path_name, base_dir=base_dir) or "")



from functools import reduce
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from hyfi.composer import BaseConfig, Composer
from hyfi.module import ModuleConfig
from hyfi.path.task import TaskPathConfig
from hyfi.pipeline.config import PipelineConfig, Pipelines, run_pipe
from hyfi.utils.contexts import change_directory, elapsed_timer
from hyfi.utils.logging import LOGGING
from hyfi.utils.packages import PKGs

logger = LOGGING.getLogger(__name__)


class TaskConfig(BaseConfig):
    _config_name_: str = "__init__"
    _config_group_: str = "/task"

    task_name: str = "demo-task"
    task_root: str = "workspace"
    version: str = "0.0.0"
    module: Optional[ModuleConfig] = None
    path: TaskPathConfig = TaskPathConfig()
    pipelines: Optional[List[Union[str, Dict]]] = []

    _property_set_methods_ = {
        "task_name": "set_task_name",
        "task_root": "set_task_root",
    }

    def set_task_root(self, val: Union[str, Path]):
        if not self.task_root or self.task_root != val:
            self.path.task_root = str(val)

    def set_task_name(self, val):
        if not self.task_name or self.task_name != val:
            self.path.task_name = val

    @property
    def config(self):
        return self.model_dump()

    @property
    def root_dir(self) -> Path:
        return self.path.root_dir

    @property
    def task_dir(self) -> Path:
        return self.path.task_dir

    @property
    def project_dir(self) -> Path:
        return self.path.project_dir

    @property
    def workspace_dir(self) -> Path:
        return self.path.workspace_dir

    @property
    def output_dir(self) -> Path:
        return self.path.output_dir

    @property
    def model_dir(self) -> Path:
        return self.path.model_dir

    @property
    def log_dir(self) -> Path:
        return self.path.log_dir

    @property
    def cache_dir(self) -> Path:
        return self.path.cache_dir

    @property
    def library_dir(self) -> Path:
        return self.path.library_dir

    @property
    def dataset_dir(self):
        return self.path.dataset_dir

    def print_config(self):
        Composer.print(self.config)

    def load_modules(self):
        """Load the modules"""
        if not self.module:
            logger.info("No module to load")
            return
        if not self.module.modules:
            logger.info("No modules to load")
            return
        library_dir = self.library_dir
        for module in self.module.modules:
            name = module.name
            libname = module.libname
            liburi = module.liburi
            specname = module.specname
            libpath = library_dir / libname
            syspath = module.get("syspath")
            if syspath is not None:
                syspath = library_dir / syspath
            PKGs.ensure_import_module(name, libpath, liburi, specname, syspath)

    def reset(self, objects=None, release_gpu_memory=True):
        """Reset the memory cache"""
        if isinstance(objects, list):
            for obj in objects:
                del obj
        if release_gpu_memory:
            from hyfi.utils.gpumon import GPUMon

            GPUMon.release_gpu_memory()

    def get_pipelines(self) -> Pipelines:
        """
        Get the list of pipelines for a task

        Args:
            task: The task to get the pipelines for

        Returns:
            A list of PipelineConfig objects
        """
        self.pipelines = self.pipelines or []
        pipelines: Pipelines = []
        for name in self.pipelines:
            if isinstance(name, str) and isinstance(getattr(self, name), dict):
                pipeline = PipelineConfig(**getattr(self, name))
                if not pipeline.name:
                    pipeline.name = name
                pipelines.append(pipeline)
        return pipelines

    def run(
        self,
        pipelines: Optional[Pipelines] = None,
    ):
        """
        Run pipelines specified in the task

        Args:
            pipelines: The pipelines to run
        """
        # Run all pipelines in the task.
        pipelines = pipelines or self.get_pipelines()
        if self.verbose:
            logger.info("Running %s pipeline(s)", len(pipelines or []))
        with elapsed_timer(format_time=True) as elapsed:
            for pipeline in pipelines:
                if self.verbose:
                    logger.info("Running pipeline: %s", pipeline.name)
                initial_object = self if pipeline.use_task_as_initial_object else None
                self.run_pipeline(pipeline, initial_object)
            # Print the elapsed time.
            if self.verbose:
                logger.info(
                    " >> elapsed time for the task with %s pipelines: %s",
                    len(pipelines or []),
                    elapsed(),
                )

    def run_pipeline(
        self,
        pipeline: Union[Dict, PipelineConfig],
        initial_object: Optional[Any] = None,
    ) -> Any:
        """
        Run a pipeline given a config

        Args:
            config: PipelineConfig to run the pipeline
            initial_obj: Object to use as initial value
            task: TaskConfig to use as task

        Returns:
            The result of the pipeline
        """
        # If config is not a PipelineConfig object it will be converted to a PipelineConfig object.
        if not isinstance(pipeline, PipelineConfig):
            pipeline = PipelineConfig(**Composer.to_dict(pipeline))
        pipes = pipeline.get_pipes()
        if (
            initial_object is None
            and pipeline.initial_object is not None
            and Composer.is_instantiatable(pipeline.initial_object)
        ):
            initial_object = Composer.instantiate(pipeline.initial_object)
        # Return initial object for the initial object
        if not pipes:
            logger.warning("No pipes specified")
            return initial_object

        pipe_names = [pipe.run for pipe in pipes]
        logger.info("Applying %s pipes: %s", len(pipe_names), pipe_names)
        # Run the task in the current directory.
        if self is None:
            self = TaskConfig()
        with elapsed_timer(format_time=True) as elapsed:
            with change_directory(self.workspace_dir):
                rst = reduce(run_pipe, pipes, initial_object)
            # Print the elapsed time.
            if pipeline.verbose:
                logger.info(
                    " >> elapsed time for the pipeline with %s pipes: %s",
                    len(pipes),
                    elapsed(),
                )
        return rst



"""Utilities for working with notebooks.""" ""
import contextlib
import logging
import os
import sys

from hyfi.utils.envs import ENVs
from hyfi.utils.iolibs import IOLIBs

logger = logging.getLogger(__name__)


class NBs:
    @staticmethod
    def is_notebook():
        """Check if the code is running in a notebook."""
        try:
            get_ipython  # type: ignore
        except NameError:
            return False
        # pylint: disable=undefined-variable
        shell_type = get_ipython().__class__.__name__  # type: ignore # noqa
        # logger.info(f"shell type: {shell_type}")
        return shell_type in ["ZMQInteractiveShell", "Shell"]

    @staticmethod
    def is_colab():
        """Check if the code is running in Google Colab."""
        is_colab = "google.colab" in sys.modules
        if is_colab:
            logger.info("Google Colab detected.")
        else:
            logger.info("Google Colab not detected.")
        return is_colab

    @staticmethod
    def get_display():
        """Get the display object for the current environment."""
        try:
            from ipywidgets import Output
        except ImportError:
            logger.info("ipywidgets not installed.")
            return None

        return Output() if NBs.is_notebook() else None

    @staticmethod
    def clear_output(wait=False):
        """Clear the output of the current notebook."""
        from IPython import display

        if NBs.is_notebook():
            display.clear_output(wait=wait)

    @staticmethod
    def display(
        *objs,
        include=None,
        exclude=None,
        metadata=None,
        transient=None,
        display_id=None,
        **kwargs,
    ):
        """Display an object in the current notebook."""
        from IPython import display

        if NBs.is_notebook() and objs is not None:
            return display.display(
                *objs,
                include=include,
                exclude=exclude,
                metadata=metadata,
                transient=transient,
                display_id=display_id,
                **kwargs,
            )

    @staticmethod
    def display_image(
        data=None,
        url=None,
        filename=None,
        format=None,
        embed=None,
        width=None,
        height=None,
        retina=False,
        unconfined=False,
        metadata=None,
        **kwargs,
    ):
        """
        Display an image, which can be given as raw data or a URL.

        Parameters
        ----------
        data : unicode, str or bytes
            The raw image data or a URL or filename to load the data from.
            This always results in embedded image data.
        url : unicode
            A URL to download the data from. If you specify `url=`,
            the image data will not be embedded unless you also specify `embed=True`.
        filename : unicode
            Path to a local file to load the data from.
            Images from a file are always embedded.
        format : unicode
            The format of the image data (png/jpeg/jpg/gif). If a filename or URL is given
            for format will be inferred from the filename extension.
        embed : bool
            Should the image data be embedded using a data URI (True) or be
            loaded using an <img> tag. Set this to True if you want the image
            to be viewable later with no internet connection in the notebook.

            Default is `True`, unless the keyword argument `url` is set, then
            default value is `False`.

            Note that QtConsole is not able to display images if `embed` is set to `False`
        width : int
            Width in pixels to which to constrain the image in html
        height : int
            Height in pixels to which to constrain the image in html
        retina : bool
            Automatically set the width and height to half of the measured
            width and height.
            This only works for embedded images because it reads the width/height
            from image data.
            For non-embedded images, you can just set the desired display width
            and height directly.
        unconfined: bool
            Set unconfined=True to disable max-width confinement of the image.
        metadata: dict
            Specify extra metadata to attach to the image.

        """
        from IPython import display

        if NBs.is_notebook():
            img = display.Image(
                data=data,
                url=url,
                filename=filename,
                format=format,
                embed=embed,
                width=width,
                height=height,
                retina=retina,
                unconfined=unconfined,
                metadata=metadata,
                **kwargs,
            )
            return display.display(img)

    @staticmethod
    def hide_code_in_slideshow():
        """Hide code in slideshow."""
        import binascii

        from IPython import display

        uid = binascii.hexlify(os.urandom(8)).decode()
        html = """<div id="%s"></div>
        <script type="text/javascript">
            $(function(){
                var p = $("#%s");
                if (p.length==0) return;
                while (!p.hasClass("cell")) {
                    p=p.parent();
                    if (p.prop("tagName") =="body") return;
                }
                var cell = p;
                cell.find(".input").addClass("hide-in-slideshow")
            });
        </script>""" % (
            uid,
            uid,
        )
        display.display_html(html, raw=True)

    @staticmethod
    def colored_str(s, color="black"):
        """Colored string."""
        # return "<text style=color:{}>{}</text>".format(color, s)
        return "<text style=color:{}>{}</text>".format(color, s.replace("\n", "<br>"))

    @staticmethod
    def cprint(str_tuples):
        from IPython.core.display import HTML as html_print
        from IPython.display import display

        display(
            html_print(
                " ".join([NBs.colored_str(ti, color=ci) for ti, ci in str_tuples])
            )
        )

    @staticmethod
    def create_dropdown(
        options, value, description, disabled=False, style=None, layout=None, **kwargs
    ):
        """Create a dropdown widget."""
        import ipywidgets as widgets

        if style is None:
            style = {"description_width": "initial"}

        layout = (
            widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
        )
        return widgets.Dropdown(
            options=options,
            value=value,
            description=description,
            disabled=disabled,
            style=style,
            layout=layout,
            **kwargs,
        )

    @staticmethod
    def create_textarea(
        value,
        description,
        placeholder="",
        disabled=False,
        style=None,
        layout=None,
        **kwargs,
    ):
        """Create a textarea widget."""
        import ipywidgets as widgets

        if style is None:
            style = {"description_width": "initial"}

        layout = (
            widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
        )
        return widgets.Textarea(
            value=value,
            placeholder=placeholder,
            description=description,
            disabled=disabled,
            style=style,
            layout=layout,
            **kwargs,
        )

    @staticmethod
    def create_button(
        description, button_style="", icon="check", layout=None, **kwargs
    ):
        """Create a button widget."""
        import ipywidgets as widgets

        layout = (
            widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
        )
        return widgets.Button(
            description=description,
            button_style=button_style,
            icon=icon,
            layout=layout,
            **kwargs,
        )

    @staticmethod
    def create_radiobutton(
        options,
        description,
        value=None,
        disabled=False,
        style=None,
        layout=None,
        **kwargs,
    ):
        """Create a radiobutton widget."""
        import ipywidgets as widgets

        if style is None:
            style = {"description_width": "initial"}

        layout = (
            widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
        )
        return widgets.RadioButtons(
            options=options,
            value=value,
            description=description,
            disabled=disabled,
            style=style,
            layout=layout,
            **kwargs,
        )

    @staticmethod
    def create_image(
        filename=None,
        format=None,
        width=None,
        height=None,
        **kwargs,
    ):
        """Create an image widget."""
        import ipywidgets as widgets

        # from urllib.request import urlopen

        if filename is None:
            url = "https://assets.entelecheia.cc/img/placeholder.png"
            # img = urlopen(url).read()
            img = IOLIBs.read(url)
            _format = "png"
        else:
            img = IOLIBs.read(filename)
            _format = format or filename.split(".")[-1]
        return widgets.Image(
            value=img,
            format=_format,
            width=width,
            height=height,
            **kwargs,
        )

    @staticmethod
    def create_floatslider(
        min=0.0,
        max=1.0,
        step=0.1,
        value=None,
        description="",
        disabled=False,
        continuous_update=False,
        orientation="horizontal",
        readout=True,
        readout_format=".1f",
        style=None,
        layout=None,
        **kwargs,
    ):
        """Create a float slider widget."""
        if style is None:
            style = {"description_width": "initial"}
        import ipywidgets as widgets

        layout = (
            widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
        )
        return widgets.FloatSlider(
            min=min,
            max=max,
            step=step,
            value=value,
            description=description,
            disabled=disabled,
            continuous_update=continuous_update,
            orientation=orientation,
            readout=readout,
            readout_format=readout_format,
            style=style,
            layout=layout,
            **kwargs,
        )

    @staticmethod
    def load_extentions(exts=None):
        """Load extentions."""
        if exts is None:
            exts = ["autotime"]
        if not NBs.is_notebook():
            return
        with contextlib.suppress(ImportError):
            from IPython.core.getipython import get_ipython

            ip = get_ipython()
            if ip is None:
                return
            try:
                loaded = ip.extension_manager.loaded
                for ext in exts:
                    if ext not in loaded:
                        ip.extentension_manager.load_extension(ext)
            except AttributeError:
                for ext in exts:
                    try:
                        ip.magic(f"load_ext {ext}")
                    except ModuleNotFoundError:
                        logger.info("Extension %s not found. Install it first.", ext)

    @staticmethod
    def set_matplotlib_formats(*formats, **kwargs):
        """Set matplotlib formats."""
        if NBs.is_notebook():
            from IPython.core.display import set_matplotlib_formats

            set_matplotlib_formats(*formats, **kwargs)

    @staticmethod
    def mount_google_drive(
        project_root: str = "",
        project_name: str = "",
        mountpoint: str = "/content/drive",
        force_remount: bool = False,
        timeout_ms: int = 120000,
    ) -> None:
        """Mount Google Drive to Colab."""
        try:
            from google.colab import drive  # type: ignore

            drive.mount(mountpoint, force_remount=force_remount, timeout_ms=timeout_ms)

            if project_root:
                if not project_root.startswith(
                    os.path.sep
                ) and not project_root.startswith(".."):
                    project_root = os.path.join(mountpoint, project_root)
                ENVs.set_osenv("HYFI_PROJECT_ROOT", project_root)
                logger.info(f"Setting HYFI_PROJECT_ROOT to {project_root}")
            if project_name:
                ENVs.set_osenv("HYFI_PROJECT_NAME", project_name)
                logger.info(f"Setting HYFI_PROJECT_NAME to {project_name}")
        except ImportError:
            logger.warning("Google Colab not detected.")



from .config import (
    BaseRunConfig,
    DataframePipeConfig,
    PipeConfig,
    PipelineConfig,
    Pipelines,
    Pipes,
    RunningCalls,
    RunningConfig,
    RunningPipelines,
    RunningSteps,
    RunningTasks,
)
from .pipeline import PIPELINEs

__all__ = [
    "BaseRunConfig",
    "DataframePipeConfig",
    "PipeConfig",
    "PipelineConfig",
    "RunningPipelines",
    "Pipelines",
    "PIPELINEs",
    "Pipes",
    "RunningCalls",
    "RunningConfig",
    "RunningTasks",
    "RunningSteps",
]



from .filter import DSFilter
from .load import DSLoad
from .process import DSProcess
from .save import DSSave
from .transform import DSTransform
from .types import DatasetDictType, DatasetLikeType, DatasetType
from .utils import DSUtils


class DATASETs(
    DSLoad,
    DSUtils,
    DSSave,
    DSFilter,
    DSTransform,
    DSProcess,
):
    def __init__(self):
        super().__init__()


__all__ = [
    "DSLoad",
    "DSUtils",
    "DatasetType",
    "DatasetDictType",
    "DatasetLikeType",
    "DATASETs",
]



from pathlib import Path

from hyfi.main.config import global_config
from hyfi.utils.logging import LOGGING

from .base import BasePathConfig

logger = LOGGING.getLogger(__name__)

__default_task_name__ = "demo-task"
__default_task_root__ = "workspace"


class TaskPathConfig(BasePathConfig):
    _config_name_: str = "__task__"

    task_name: str = __default_task_name__
    task_root: str = __default_task_root__

    @property
    def name(self) -> str:
        """
        Returns the name of the path configuration.
        """
        return self.task_name

    @property
    def project_dir(self) -> Path:
        return global_config.project_dir

    @property
    def project_workspace_dir(self) -> Path:
        return global_config.project_workspace_dir

    @property
    def root_dir(self) -> Path:
        """
        Returns the path to the task root directory.

        Returns:
            an path to the task root directory or None if it doesn't exist or cannot be converted to a path object
        """
        # return as an path
        if self.task_root == __default_task_root__:
            path_ = self.project_workspace_dir
        else:
            path_ = Path(self.task_root)
            if not path_.is_absolute():
                path_ = self.project_dir / path_
        path_.mkdir(parents=True, exist_ok=True)
        return path_

    @property
    def task_dir(self) -> Path:
        """
        Returns the path to the task root directory.

        Returns:
            an path to the task root directory or None if it doesn't exist or cannot be converted to a path object
        """
        # return as an path
        path_ = self.root_dir / self.task_name
        path_.mkdir(parents=True, exist_ok=True)
        return path_

    @property
    def workspace_dir(self) -> Path:
        """
        Returns the path to the task root directory.

        Returns:
            an path to the task root directory or None if it doesn't exist or cannot be converted to a path object
        """
        return self.root_dir
